# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

from typing import Dict, List

import torch

from .dip import diptest


def find_events(
    patterns: torch.Tensor, dt_pval: float
) -> List[Dict[str, torch.Tensor]]:
    """
    Select events from the set of patterns obtained by decomposition
    of the activity scores.

    :param patterns: Tensor of shape (p, T), with $p$ the number of patterns
        and $T$ the pattern length.
    :type patterns: torch.Tensor
    :param dt_pval: P-value threshold for event detection. Passed to the dip
        test.
    :type dt_pval: float
    :return: List of detected events. Elements of the list are
        dictionaries with keys 'range' and 'event'; 'range' gives the interval
        over which the event occurred, and 'event' is a length-T array of the
        event intensity.
    :rtype: List[int, Dict[str, torch.Tensor]]
    """

    candidates = find_unimodal_events(patterns, dt_pval)
    events = merge_events(candidates)
    return events


def find_unimodal_events(patterns: torch.Tensor, dt_pval: float) -> torch.Tensor:
    """
    Given a set of patterns, return those that are unimodal. Unimodality is
    evaluated via the dip test.

    :param patterns: Tensor of shape (p, T), with $p$ the number of patterns
        and $T$ the pattern length.
    :type patterns: torch.Tensor
    :param dt_pval: P-value threshold for event detection. Passed to the dip
        test.
    :type dt_pval: float
    :return: Tensor of shape (p_u, T), with $p_u$ the number of unimodal
        patterns detected.
    :rtype: torch.Tensor
    """
    E = patterns.shape[0]
    dt_pvals = torch.zeros(E)

    for e in range(E):
        _, pval = diptest(patterns[e, :])
        dt_pvals[e] = pval

    return patterns[dt_pvals > dt_pval, :]


def merge_events(
    candidates: torch.Tensor,
) -> List[Dict[str, torch.Tensor]]:
    """
    Merge candidate events if they have sufficient overlap. Returns a
    structured object containing all final events, with information on
    their duration and intensity.

    :param candidates: Tensor of shape (p, T), with $p$ the number of
        candidate (i.e. unimodal) patterns and $T$ the pattern length.
    :type candidates: torch.Tensor
    :return: List of detected events. Elements of the list are
        dictionaries with keys 'event_window' and 'event_pattern';
        'event_window' gives the interval over which the event occurred,
        and 'event_pattern' is a length-T array of the event intensity.
    :rtype: List[int, Dict[str, torch.Tensor]]
    """
    E, T = candidates.shape

    starts = torch.zeros(E)
    ends = torch.zeros(E)
    for e in range(E):  # define event start and end by cumsum quantiles
        event = candidates[e, :]
        ecdf = torch.cumsum(event, dim=0) / torch.sum(event)
        starts[e] = torch.sum(ecdf < 0.1)
        ends[e] = torch.sum(ecdf < 0.9)
    ix = torch.argsort(starts)

    sstart = starts[ix]
    send = ends[ix]
    sevents = candidates[ix, :]

    merged: List[Dict[str, torch.Tensor]] = (
        []
    )  # merge in linear pass over time dimension
    currstart = torch.tensor([-1])
    currend = torch.tensor([-1])
    currevent = torch.ones(T) * -1.0

    for i, s in enumerate(sstart):
        if i == 0:  # first event
            currstart = s
            currend = send[i]
            currevent = sevents[i, :]
        elif s > currend:  # start of new event; record previous one
            merged.append(
                {
                    "event_window": torch.stack([currstart, currend]),
                    "event_pattern": currevent,
                }
            )
            currstart = s
            currend = send[i]
            currevent = sevents[i, :]
        else:  # update current event
            currend = torch.max(torch.stack([currend, send[i]]))
            currevent += sevents[i, :]

    # record final event
    if currstart > 0:
        merged.append(
            {
                "event_window": torch.stack([currstart, currend]),
                "event_pattern": currevent,
            }
        )
    return merged


def assign_metrics_to_events(
    events: List[Dict[str, torch.Tensor]],
    activity_scores: torch.Tensor,
    omp_tol: float,
) -> List[Dict[str, torch.Tensor]]:
    """
    Given a set of activity scores and the events detected, assign each
    activity score (i.e. metric) to an event. Each metric can be assigned
    to zero, one, or multiple events.


    :param events: Events detected in the activity score data.
    :type events: List[int, Dict[str, torch.Tensor]]
    :param activity_scores: The activiy scores from which the events were
        computed.
    :type activity_scores: torch.Tensor
    :param omp_tol: Threshold for event assignment via orthogonal matching
        pursuit. Each next event is assigned to the metric if it explains a
        proportion of the residual variance at least equal to `omp_tol`.
    :type omp_tol: float
    :return: A list with the same structure as the `events` parameter,
        where each element (event) now also has a 'metrics' field listing the
        indices of the metrics assigned to that event.
    :rtype: List[int, Dict[str, torch.Tensor]]
    """
    (
        M,
        T,
    ) = activity_scores.shape  # loop over metrics, getting their event assignments

    massign = {m: omp_assign(events, activity_scores[m, :], omp_tol) for m in range(M)}

    # invert this to get list of metrics per event
    evs: List[int] = []
    for v in massign.values():
        for ev in v:
            if ev not in evs:
                evs += [ev]

    for e in evs:
        mlist: List[int] = []
        for k, v in massign.items():
            if int(e) in v:
                mlist += [k]

        events[int(e)]["suspected_metrics"] = torch.tensor(mlist)

    return events


def omp_assign(
    events: List[Dict[str, torch.Tensor]],
    activity_score: torch.Tensor,
    omp_tol: float,
) -> List[int]:
    """
    Orthogonal matching pursuit for event assignment. Assignment is defined in
    terms of explanation: a metric is assigned to an event if that event
    explains a significant fraction of the metric's activity score. For each
    metric, the set of explanatory events is determined by orthogonal matching
    pursuit regression of the activity score against the set of event
    intensities.

    :param events: The events used to explain a metric's activity score.
        Assumed to have the same structure as the output of `find_events()`.
    :type events: dict[int, dict[str, torch.Tensor]]
    :param activity_score: 1-D activity score of the metric.
    :type activity_score: torch.Tensor
    :param omp_tol: Threshold for event assignment via orthogonal matching
        pursuit. Each next event is assigned to the metric if it explains a
        proportion of the residual variance at least equal to `omp_tol`.
    :type omp_tol: float
    :return: List of events assigned to the metric.
    :rtype: List[int]
    """

    if torch.var(activity_score, dim=0, unbiased=False) == 0:
        return []  # skip assignment if activity score is flat

    # initialize for OMP loop
    assign: List[int] = []
    assigned_events = torch.empty((len(activity_score), 0))
    cvg = False
    events_left: List[int] = torch.arange(len(events)).tolist()
    r = activity_score

    if len(events_left) == 0:  # case where no events were found upstream
        return assign

    while not cvg:
        # get remaining event patterns
        event_mat = torch.stack(
            [events[i]["event_pattern"] for i in events_left], dim=1
        )
        event_mat /= torch.linalg.norm(event_mat, dim=0)

        # find event that best explains activity score remainder
        corrs = torch.matmul(event_mat.t(), r)
        eix = int(torch.argmax(torch.abs(corrs)))

        # add it if it sufficiently reduces approximation error
        assigned_events_trial = torch.cat(
            (assigned_events, event_mat[:, eix].unsqueeze(1)), dim=1
        )
        coef = torch.linalg.solve(
            torch.mm(assigned_events_trial.t(), assigned_events_trial),
            torch.matmul(assigned_events_trial.t(), activity_score),
        )
        rnew = activity_score - torch.matmul(assigned_events_trial, coef)

        rel_sqerr_decrease = torch.sum(rnew**2) / torch.sum(r**2)
        if rel_sqerr_decrease < omp_tol:
            r = rnew
            assign += [events_left[eix]]
            assigned_events = assigned_events_trial
            events_left.pop(eix)

            if len(events_left) == 0:
                cvg = True  # end if no events remain

        else:
            cvg = True  # no remaining events sufficiently explain data

    return assign



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json

from opensearchpy import OpenSearch

from opensearch_py_ml.ml_commons.ml_common_utils import ML_BASE_URI


class ModelExecute:
    """
    Class for executing algorithms using ML Commons execute API.
    """

    API_ENDPOINT = "models/_execute"

    def __init__(self, os_client: OpenSearch):
        self._client = os_client

    def _execute(self, algorithm_name: str, input_json: dict) -> dict:
        """
        This method executes ML algorithms that can be only executed directly (i.e. do not support train and
        predict APIs), like anomaly localization and metrics correlation.
        The input json must be a dictionary or a deserializable Python object.
        The algorithm has to be supported by ML Commons.
        Refer to https://opensearch.org/docs/2.7/ml-commons-plugin/api/#execute

        :param algorithm_name: Name of the algorithm
        :type algorithm_name: string
        :param input_json: Dictionary of parameters or a deserializable string, byte, or bytearray
        :type input_json: dict
        :return: returns the API response
        :rtype: dict
        """

        if not isinstance(input_json, dict):
            input_json = json.loads(input_json)

        return self._client.transport.perform_request(
            method="POST",
            url=f"{ML_BASE_URI}/_execute/{algorithm_name}",
            body=input_json,
        )



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
import math
import os
import pickle
import platform
import random
import re
import shutil
import subprocess
import time
from pathlib import Path
from typing import List
from zipfile import ZipFile

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import yaml
from accelerate import Accelerator, notebook_launcher
from mdutils.fileutils import MarkDownFile
from sentence_transformers import SentenceTransformer
from sentence_transformers.models import Normalize, Pooling, Transformer
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import TrainingArguments, get_linear_schedule_with_warmup
from transformers.convert_graph_to_onnx import convert
from transformers.models.distilbert.modeling_distilbert import (
    DistilBertSdpaAttention,
    MultiHeadSelfAttention,
)

from opensearch_py_ml.ml_commons.ml_common_utils import (
    _generate_model_content_hash_value,
)

from .base_models import BaseUploadModel


class SentenceTransformerModel(BaseUploadModel):
    """
    Class for training, exporting and configuring the SentenceTransformers model.
    """

    DEFAULT_MODEL_ID = "sentence-transformers/msmarco-distilbert-base-tas-b"
    SYNTHETIC_QUERY_FOLDER = "synthetic_queries"

    MODEL_SPACE_TYPE_MAPPING = {
        "all-distilroberta-v1": "l2",
        "all-MiniLM-L6-v2": "l2",
        "all-MiniLM-L12-v2": "l2",
        "all-mpnet-base-v2": "l2",
        "msmarco-distilbert-base-tas-b": "innerproduct",
        "multi-qa-MiniLM-L6-cos-v1": "l2",
        "multi-qa-mpnet-base-dot-v1": "innerproduct",
        "paraphrase-MiniLM-L3-v2": "cosinesimil",
        "paraphrase-multilingual-MiniLM-L12-v2": "cosinesimil",
        "paraphrase-mpnet-base-v2": "cosinesimil",
        "distiluse-base-multilingual-cased-v1": "cosinesimil",
    }

    def __init__(
        self,
        model_id: str = DEFAULT_MODEL_ID,
        folder_path: str = None,
        overwrite: bool = False,
    ) -> None:
        """
        Initiate a sentence transformer model class object. The model id will be used to download
        pretrained model from the hugging-face and served as the default name for model files, and the folder_path
        will be the default location to store files generated in the following functions

        :param model_id: Optional, the huggingface mode id to download sentence transformer model,
            default model id: 'sentence-transformers/msmarco-distilbert-base-tas-b'
        :type model_id: string
        :param folder_path: Optional, the path of the folder to save output files, such as queries, pre-trained model,
            after-trained custom model and configuration files. if None, default as "/model_files/" under the current
            work directory
        :type folder_path: string
        :param overwrite: Optional,  choose to overwrite the folder at folder path. Default as false. When training
                    different sentence transformer models, it's recommended to give designated folder path every time.
                    Users can choose to overwrite = True to overwrite previous runs
        :type overwrite: bool
        :return: no return value expected
        :rtype: None
        """
        super().__init__(model_id, folder_path, overwrite)
        default_folder_path = os.path.join(
            os.getcwd(), "sentence_transformer_model_files"
        )

        if folder_path is None:
            self.folder_path = default_folder_path
        else:
            self.folder_path = folder_path

        # Check if self.folder_path exists
        if os.path.exists(self.folder_path) and not overwrite:
            print(
                "To prevent overwriting, please enter a different folder path or delete the folder or enable "
                "overwrite = True "
            )
            raise Exception(
                str("The default folder path already exists at : " + self.folder_path)
            )

        self.model_id = model_id
        self.torch_script_zip_file_path = None
        self.onnx_zip_file_path = None

    def train(
        self,
        read_path: str,
        overwrite: bool = False,
        output_model_name: str = None,
        zip_file_name: str = None,
        compute_environment: str = None,
        num_machines: int = 1,
        num_gpu: int = 0,
        learning_rate: float = 2e-5,
        num_epochs: int = 10,
        batch_size: int = 32,
        verbose: bool = False,
        percentile: float = 95,
    ) -> None:
        """
        Read the synthetic queries and use it to fine tune/train (and save) a sentence transformer model.

        Parameters
        ----------
        :param read_path:
            required, path to the zipped file that contains generated queries, if None, raise exception.
            the zipped file should contain pickled file in list of dictionary format with key named as 'query',
            'probability' and 'passages'. For example: [{'query':q1,'probability': p1,'passages': pa1}, ...].
            'probability' is not required for training purpose
        :type read_path: string
        :param overwrite:
            optional, synthetic_queries/ folder in current directory is to store unzip queries files.
            Default to set overwrite as false and if the folder is not empty, raise exception to recommend users
            to either clean up folder or enable overwriting is True
        :type overwrite: bool
        :param output_model_name:
            the name of the trained custom model. If None, default as model_id + '.pt'
        :type output_model_name: string
        :param zip_file_name:
            Optional, file name for zip file. if None, default as model_id + '.zip'
        :type zip_file_name: string
        :param compute_environment:
            optional, compute environment type to run model, if None, default using `LOCAL_MACHINE`
        :type compute_environment: string
        :param num_machines:
            optional, number of machine to run model , if None, default using 1
        :type num_machines: int
        :param num_gpu:
            optional, number of gpus to run model , if None, default to 0. If number of gpus > 1, use HuggingFace
            accelerate to launch distributed training
        :param learning_rate:
            optional, learning rate to train model, default is 2e-5
        :type learning_rate: float
        :param num_epochs:
            optional, number of epochs to train model, default is 10
        :type num_epochs: int
        :param batch_size:
            optional, batch size for training, default is 32
        :type batch_size: int
        :param verbose:
            optional, use plotting to plot the training progress. Default as false
        :type verbose: bool
        :param percentile:
            we find the max length of {percentile}% of the documents. Default is 95%
            Since this length is measured in terms of words and not tokens we multiply it by 1.4 to approximate the fact
            that 1 word in the english vocabulary roughly translates to 1.3 to 1.5 tokens
        :type percentile: float

        Returns
        -------
        :return: no return value expected
        :rtype: None
        """

        query_df = self.read_queries(read_path, overwrite)

        train_examples = self.load_training_data(query_df)

        if num_gpu > 1:
            self.set_up_accelerate_config(
                compute_environment=compute_environment,
                num_machines=num_machines,
                num_processes=num_gpu,
                verbose=verbose,
            )

            if self.__is_notebook():
                # MPS needs to be only enabled for MACOS: https://pytorch.org/docs/master/notes/mps.html
                if platform.system() == "Darwin":
                    if not torch.backends.mps.is_available():
                        if not torch.backends.mps.is_built():
                            print(
                                "MPS not available because the current PyTorch install was not "
                                "built with MPS enabled."
                            )
                        else:
                            print(
                                "MPS not available because the current MacOS version is not 12.3+ "
                                "and/or you do not have an MPS-enabled device on this machine."
                            )
                        exit(1)  # Existing the script as the script will break anyway
                notebook_launcher(
                    self.train_model,
                    args=(
                        train_examples,
                        self.model_id,
                        output_model_name,
                        learning_rate,
                        num_epochs,
                        batch_size,
                        verbose,
                        num_gpu,
                        percentile,
                    ),
                    num_processes=num_gpu,
                )
            else:
                try:
                    subprocess.run(
                        [
                            "accelerate",
                            "launch",
                            self.train_model(
                                train_examples,
                                self.model_id,
                                output_model_name,
                                learning_rate,
                                num_epochs,
                                batch_size,
                                verbose,
                                num_gpu,
                                percentile,
                            ),
                        ],
                    )
                # TypeError: expected str, bytes or os.PathLike object, not TopLevelTracedModule happens after
                # running process.
                except TypeError:
                    self.zip_model(self.folder_path, output_model_name, zip_file_name)
                    return None

        else:  # Do not use accelerate when num_gpu is 1 or 0
            self.train_model(
                train_examples,
                self.model_id,
                output_model_name,
                learning_rate,
                num_epochs,
                batch_size,
                verbose,
                num_gpu,
                percentile,
            )

        self.zip_model(self.folder_path, output_model_name, zip_file_name)
        return None

    #    public step by step functions:
    def read_queries(self, read_path: str, overwrite: bool = False) -> pd.DataFrame:
        """
        Read the queries generated from the Synthetic Query Generator (SQG) model, unzip files to current directory
        within synthetic_queries/ folder, output as a dataframe

        :param read_path:
            required, path to the zipped file that contains generated queries
        :type read_path: string
        :param overwrite:
            optional, synthetic_queries/ folder in current directory is to store unzip queries files.
            Default to set overwrite as false and if the folder is not empty, raise exception to recommend users
            to either clean up folder or enable overwriting is True
        :type overwrite: bool
        :return: The dataframe of queries.
        :rtype: panda dataframe
        """
        # assign a local folder 'synthetic_queries/' to store the unzip file,
        # check if the folder contains sub-folders and files, remove and clean up the folder before unzip.
        # walk through the zip file and read the file paths into file_list
        unzip_path = os.path.join(self.folder_path, self.SYNTHETIC_QUERY_FOLDER)

        if os.path.exists(unzip_path):
            if len(os.listdir(unzip_path)) > 0:
                if overwrite:
                    for files in os.listdir(unzip_path):
                        sub_path = os.path.join(unzip_path, files)
                        if os.path.isfile(sub_path):
                            os.remove(sub_path)
                        else:
                            try:
                                shutil.rmtree(sub_path)
                            except OSError as err:
                                print(
                                    "Failed to delete files, please delete all files in "
                                    + str(unzip_path)
                                    + " "
                                    + str(err)
                                )
                else:
                    raise Exception(
                        "'synthetic_queries' folder is not empty, please clean up folder, or enable overwrite = "
                        + "True. Try again. Please check "
                        + unzip_path
                    )

        # appending all the file paths of synthetic query files in a list.
        file_list = []
        process = []
        with ZipFile(read_path, "r") as zip_ref:
            zip_ref.extractall(unzip_path)

        for root, dirnames, filenames in os.walk(unzip_path):
            for filename in filenames:
                file_list.append(os.path.join(root, filename))

        # check empty zip file
        num_file = len(file_list)

        if num_file == 0:
            raise Exception(
                "Zipped file is empty. Please provide a zip file with synthetic queries."
            )

        for file_path in file_list:
            try:
                with open(file_path, "rb") as f:
                    print("Reading synthetic query file: " + file_path + "\n")
                    process.append(pickle.load(f))
            except IOError:
                print("Failed to open synthetic query file: " + file_path + "\n")

        # reading the files to get the probability, queries and passages
        prob = []
        query = []
        passages = []
        for j in range(0, num_file):
            for dict_str in process[j]:
                if "query" in dict_str.keys() and "passage" in dict_str.keys():
                    query.append(dict_str["query"])
                    passages.append(dict_str["passage"])
                    if "probability" in dict_str.keys():
                        prob.append(dict_str["probability"])  # language modeling score
                    else:
                        prob.append(
                            "-1"
                        )  # "-1" will serve as a label saying that the probability does not exist.

        df = pd.DataFrame(
            list(zip(prob, query, passages)), columns=["prob", "query", "passages"]
        )

        # dropping duplicate queries
        df = df.drop_duplicates(subset=["query"])
        # for removing the "QRY:" token if they exist in passages
        df["passages"] = df.apply(lambda x: self.__qryrem(x), axis=1)
        # shuffle data within dataframe
        df = df.sample(frac=1)
        return df

    def load_training_data(self, query_df) -> List[List[str]]:
        """
        Create input data for training the model

        :param query_df:
            required for loading training data
        :type query_df: pd.DataFrame
        :return: the list of train examples.
        :rtype: list
        """

        train_examples = []
        print("Loading training examples... \n")
        queries = list(query_df["query"])
        passages = list(query_df["passages"])
        for i in tqdm(range(len(query_df)), total=len(query_df)):
            train_examples.append([queries[i], passages[i]])
        return train_examples

    def _get_parent_and_attr(self, model, module_name):
        """
        Retrieve the parent module and the attribute name for a given module.
        For example, if module_name is "encoder.layer.0.attention", this function will return
        the "encoder.layer.0" module and "attention" as the attribute name.

        :param model:
            required, the model instance to traverse
        :type model: SentenceTransformer
        :param module_name:
            required, the dot-separated path to the module
        :type module_name: string
        :return: tuple containing parent module object and the name of the target attribute
        :rtype: Tuple[object, string]
        """
        parts = module_name.split(".")
        parent = model
        for part in parts[:-1]:  # Traverse until the second last part
            parent = getattr(parent, part)
        return parent, parts[-1]

    def patch_model_weights(self, model):
        """
        Replace DistilBertSdpaAttention with MultiHeadSelfAttention in the given model.
        Needed for compatibility with newer PyTorch version (2.5.1) while preserving the model's learned weights.

        This function performs the following steps:
        1. Identifies all DistilBertSdpaAttention layers in the model
        2. Creates new MultiHeadSelfAttention layers with identical weights and configuration
        3. Replaces the old DistilBertSdpaAttention attention layers with the new MultiHeadSelfAttention attention layers
        4. Ensures the forward method returns proper tuple format for compatibility

        :param model:
            required, the DistilBert model instance
        :type model: SentenceTransformer
        :return: the modified model with replaced attention layers
        :rtype: SentenceTransformer
        """
        # Collect the layers to replace in a separate list to avoid modifying dictionary while iterating
        modules_to_replace = []

        for name, module in model.named_modules():
            # Identify modules that need to be replaced
            if isinstance(module, DistilBertSdpaAttention):
                modules_to_replace.append((name, module))

        # Replace the identified modules
        for name, module in modules_to_replace:
            # Retrieve the original config
            config = getattr(module, "config", None)
            if config is None:
                raise ValueError(f"Module {name} does not have a 'config' attribute.")

            # Create new MultiHeadSelfAttention with same config
            new_module = MultiHeadSelfAttention(config)

            # Copy weights into new module
            new_module.q_lin.weight.data = module.q_lin.weight.data.clone()
            new_module.q_lin.bias.data = module.q_lin.bias.data.clone()
            new_module.k_lin.weight.data = module.k_lin.weight.data.clone()
            new_module.k_lin.bias.data = module.k_lin.bias.data.clone()
            new_module.v_lin.weight.data = module.v_lin.weight.data.clone()
            new_module.v_lin.bias.data = module.v_lin.bias.data.clone()
            new_module.out_lin.weight.data = module.out_lin.weight.data.clone()
            new_module.out_lin.bias.data = module.out_lin.bias.data.clone()

            # Modify the forward method to fix tuple return issue
            def new_forward(
                self, query, key, value, mask, head_mask, output_attentions
            ):
                """
                Replace forward method to ensure proper tuple return format.

                :param query:
                    required, tensor containing query vectors of shape [batch_size, seq_length, dim]
                :type query: Tensor
                :param key:
                    required, tensor containing key vectors of shape [batch_size, seq_length, dim]
                :type key: Tensor
                :param value:
                    required, tensor containing value vectors of shape [batch_size, seq_length, dim]
                :type value: Tensor
                :param mask:
                    required, attention mask tensor of shape [batch_size, seq_length] or [batch_size, seq_length, seq_length]
                :type mask: Tensor
                :param head_mask:
                    required, mask for attention heads
                :type head_mask: Tensor
                :param output_attentions:
                    required, boolean flag indicating whether to output attention weights
                :type output_attentions: bool
                :return: A tuple of output and the attention weights if output_attentions is True,
                        otherwise a tuple of output. Output shape: [batch_size, seq_length, dim]
                :rtype: Tuple[Tensor, Tensor] or Tuple[Tensor]
                """
                batch_size, seq_length, _ = query.shape
                dim_per_head = self.dim // self.n_heads

                # Ensure the mask is the correct shape
                if mask.dim() == 2:  # [batch_size, seq_length]
                    mask = mask[
                        :, None, None, :
                    ]  # Convert to [batch_size, 1, 1, seq_length]
                elif mask.dim() == 3:  # [batch_size, seq_length, seq_length]
                    mask = mask[
                        :, None, :, :
                    ]  # Convert to [batch_size, 1, seq_length, seq_length]

                # Validate the new mask shape before applying expansion
                if mask.shape[-1] != seq_length:
                    raise ValueError(
                        f"Mask shape {mask.shape} does not match sequence length {seq_length}"
                    )

                # Apply mask expansion for all attention heads
                mask = (mask == 0).expand(
                    batch_size, self.n_heads, seq_length, seq_length
                )

                # Transform query, key, and value for multi-head attention
                q = (
                    self.q_lin(query)
                    .view(batch_size, seq_length, self.n_heads, dim_per_head)
                    .transpose(1, 2)
                )
                k = (
                    self.k_lin(key)
                    .view(batch_size, seq_length, self.n_heads, dim_per_head)
                    .transpose(1, 2)
                )
                v = (
                    self.v_lin(value)
                    .view(batch_size, seq_length, self.n_heads, dim_per_head)
                    .transpose(1, 2)
                )

                q = q / math.sqrt(dim_per_head)
                scores = torch.matmul(q, k.transpose(-2, -1))

                # Apply the correctly shaped mask
                scores = scores.masked_fill(mask, torch.finfo(scores.dtype).min)

                weights = nn.functional.softmax(scores, dim=-1)
                weights = nn.functional.dropout(
                    weights, p=self.dropout.p, training=self.training
                )

                context = torch.matmul(weights, v)
                context = (
                    context.transpose(1, 2)
                    .contiguous()
                    .view(batch_size, seq_length, self.dim)
                )
                output = self.out_lin(context)

                # Ensure return is always a tuple, as expected by DistilBERT
                return (output, weights) if output_attentions else (output,)

            # Replace forward method with the new function
            new_module.forward = new_forward.__get__(new_module, MultiHeadSelfAttention)

            # Replace module in the model
            parent_module, attr_name = self._get_parent_and_attr(model, name)
            setattr(parent_module, attr_name, new_module)

        return model

    def train_model(
        self,
        train_examples: List[List[str]],
        model_id: str = None,
        output_model_name: str = None,
        learning_rate: float = 2e-5,
        num_epochs: int = 10,
        batch_size: int = 32,
        verbose: bool = False,
        num_gpu: int = 0,
        percentile: float = 95,
    ):
        """
        Takes in training data and a sentence transformer url to train a custom semantic search model

        :param train_examples:
            required, input for the sentence transformer model training
        :type train_examples: List of strings in another list
        :param model_id:
            [optional] the url to download sentence transformer model, if None,
            default as 'sentence-transformers/msmarco-distilbert-base-tas-b'
        :type model_id: string
        :param output_model_name:
            optional,the name of the trained custom model. If None, default as model_id + '.pt'
        :type output_model_name: string
        :param learning_rate:
            optional, learning rate to train model, default is 2e-5
        :type learning_rate: float
        :param num_epochs:
            optional, number of epochs to train model, default is 10
        :type num_epochs: int
        :param batch_size:
            optional, batch size for training, default is 32
        :type batch_size: int
        :param verbose:
            optional, use plotting to plot the training progress and printing more logs. Default as false
        :type verbose: bool
        :param num_gpu:
            Number of gpu will be used for training. Default 0
        :type num_gpu: int
        :param percentile:
            To save memory while training we truncate all passages beyond a certain max_length.
            Most middle-sized transformers have a max length limit of 512 tokens. However, certain corpora can
            have shorter documents. We find the word length of all documents, sort them in increasing order and
            take the max length of {percentile}% of the documents. Default is 95%
        :type percentile: float
        :return: the torch script format trained model.
        :rtype: .pt file
        """

        if model_id is None:
            model_id = "sentence-transformers/msmarco-distilbert-base-tas-b"
        if output_model_name is None:
            output_model_name = str(self.model_id.split("/")[-1] + ".pt")

        # declare variables before assignment for training
        corp_len = []

        # Load a model from HuggingFace
        model = SentenceTransformer(model_id)

        # Calculate the length of passages
        for i in range(len(train_examples)):
            corp_len.append(len(train_examples[i][1].split(" ")))

        # In the following, we find the max length of 95% of the documents (when sorted by increasing word length).
        # Since this length is measured in terms of words and not tokens we multiply it by 1.4 to approximate the
        # fact that 1 word in the english vocabulary roughly translates to 1.3 to 1.5 tokens. For instance the word
        # butterfly will be split by most tokenizers into butter and fly, but the word sun will be probably kept as it
        # is. Note that this ratio will be higher if the corpus is jargon heavy and/or domain specific.

        corp_max_tok_len = int(np.percentile(corp_len, percentile) * 1.4)
        model.tokenizer.model_max_length = corp_max_tok_len
        model.max_seq_length = corp_max_tok_len

        # use accelerator for training
        if num_gpu > 1:
            # the default_args are required for initializing train_dataloader,
            # but output_dir is not used in this function.
            default_args = {
                "output_dir": "~/",
                "evaluation_strategy": "steps",
                "num_train_epochs": num_epochs,
                "log_level": "error",
                "report_to": "none",
            }
            training_args = TrainingArguments(
                per_device_train_batch_size=batch_size,
                gradient_accumulation_steps=1,
                fp16=False,
                **default_args,
            )

            train_dataloader = DataLoader(
                train_examples,
                shuffle=True,
                batch_size=training_args.per_device_train_batch_size,  # Trains with this batch size.
            )

            optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate)
            scheduler = get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=min(10000, 0.05 * len(train_dataloader)),
                num_training_steps=num_epochs * len(train_dataloader),
            )

            accelerator = Accelerator()
            model, optimizer, train_dataloader, scheduler = accelerator.prepare(
                model, optimizer, train_dataloader, scheduler
            )
            print("Device using for training: ", accelerator.device)
            model.to(accelerator.device)
            init_time = time.time()
            total_loss = []

            if accelerator.process_index == 0:
                print("Start training with accelerator...\n")

                print(f"The number of training epochs per process are {num_epochs}\n")

                print(
                    f"The total number of steps per training epoch are {len(train_dataloader)}\n"
                )

            # The following training loop trains the sentence transformer model using the standard contrastive loss
            # with in-batch negatives. The particular contrastive loss that we use is the Multi-class N-pair Loss (
            # eq. 6, Sohn, NeurIPS 2016). In addition, we symmetrize the loss with respect to queries and passages (
            # as also used in OpenAI's CLIP model). The performance improves with the number of in-batch negatives
            # but larger batch sizes can lead to out of memory issues, so please use the batch-size judiciously.

            for epoch in range(num_epochs):
                print(
                    "Training epoch "
                    + str(epoch)
                    + " in process "
                    + str(accelerator.process_index)
                    + "...\n"
                )
                for step, batch in tqdm(
                    enumerate(train_dataloader), total=len(train_dataloader)
                ):
                    batch_q = batch[0]
                    batch_p = batch[1]

                    model = accelerator.unwrap_model(model)
                    out_q = accelerator.unwrap_model(model).tokenize(batch_q)
                    for key in out_q.keys():
                        out_q[key] = out_q[key].to(model.device)

                    out_p = accelerator.unwrap_model(model).tokenize(batch_p)
                    for key in out_p.keys():
                        out_p[key] = out_p[key].to(model.device)

                    Y = model(out_q)["sentence_embedding"]
                    X = model(out_p)["sentence_embedding"]

                    XY = torch.exp(torch.matmul(X, Y.T) / (X.shape[1]) ** 0.5)
                    num = torch.diagonal(XY)

                    den0 = torch.sum(XY, dim=0)
                    den1 = torch.sum(XY, dim=1)

                    batch_loss = -torch.sum(torch.log(num / den0)) - torch.sum(
                        torch.log(num / den1)
                    )

                    accelerator.backward(batch_loss)
                    optimizer.step()
                    scheduler.step()
                    optimizer.zero_grad()
                    total_loss.append(batch_loss.item())

                    if verbose is True and not step % 500 and step != 0:
                        plt.plot(total_loss[::100])
                        plt.show()
            accelerator.wait_for_everyone()

        # When number of GPU is less than 2, we don't need to accelerate
        else:
            # identify if running on gpu or cpu
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            model.to(device)

            total_steps = (len(train_examples) // batch_size) * num_epochs
            steps_size = len(train_examples) // batch_size
            optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate)
            scheduler = get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=max(10000, total_steps * 0.05),
                num_training_steps=total_steps,
            )

            loss = []
            init_time = time.time()

            print("Start training without accelerator...\n")
            print(f"The number of training epoch are {num_epochs}\n")
            print(
                f"The total number of steps per training epoch are {len(train_examples) // batch_size}\n"
            )

            # The following training loop trains the sentence transformer model using the standard contrastive loss
            # with in-batch negatives. The particular contrastive loss that we use is the Multi-class N-pair Loss (
            # eq. 6, Sohn, NeurIPS 2016). In addition, we symmetrize the loss with respect to queries and passages (
            # as also used in OpenAI's CLIP model). The performance improves with the number of in-batch negatives
            # but larger batch sizes can lead to out of memory issues, so please use the batch-size judiciously.

            for epoch in range(num_epochs):
                random.shuffle(train_examples)
                print("Training epoch " + str(epoch) + "...\n")
                for j in tqdm(range(steps_size), total=steps_size):
                    batch_q = []
                    batch_p = []
                    for example in train_examples[
                        j * batch_size : (j + 1) * batch_size
                    ]:
                        batch_q.append(example[0])
                        batch_p.append(example[1])

                    out_q = model.tokenize(batch_q)
                    for key in out_q.keys():
                        out_q[key] = out_q[key].to(device)
                    Y = model(out_q)["sentence_embedding"]

                    out = model.tokenize(batch_p)
                    for key in out.keys():
                        out[key] = out[key].to(device)
                    X = model(out)["sentence_embedding"]

                    XY = torch.exp(torch.matmul(X, Y.T) / (X.shape[1]) ** 0.5)
                    num = torch.diagonal(XY)

                    den0 = torch.sum(XY, dim=0)
                    den1 = torch.sum(XY, dim=1)

                    train_loss = -torch.sum(torch.log(num / den0)) - torch.sum(
                        torch.log(num / den1)
                    )
                    loss.append(train_loss.item())
                    train_loss.backward()
                    optimizer.step()
                    scheduler.step()
                    optimizer.zero_grad()

                    if verbose is True and not j % 500 and j != 0:
                        plt.plot(loss[::100])
                        plt.show()

        model = self.patch_model_weights(model)
        # saving the pytorch model and the tokenizers.json file is saving at this step
        model.save(self.folder_path)
        device = "cpu"
        cpu_model = model.to(device)
        print(f"Total training time: {time.time() - init_time}\n")

        for key in out_q.keys():
            out_q[key] = out_q[key].to(device)

        traced_cpu = torch.jit.trace(
            cpu_model,
            (
                {
                    "input_ids": out_q["input_ids"],
                    "attention_mask": out_q["attention_mask"],
                }
            ),
            strict=False,
        )
        if verbose:
            print("Preparing model to save...\n")
        torch.jit.save(traced_cpu, os.path.join(self.folder_path, output_model_name))
        print("Model saved to path: " + self.folder_path + "\n")
        return traced_cpu

    def zip_model(
        self,
        model_path: str = None,
        model_name: str = None,
        zip_file_name: str = None,
        add_apache_license: bool = False,
        verbose: bool = False,
    ) -> None:
        """
        Zip the model file and its tokenizer.json file to prepare to upload to the OpenSearch cluster

        :param model_path:
            Optional, path to find the model file, if None, default as concatenate model_id and
            '.pt' file in current path
        :type model_path: string
        :param model_name:
            the name of the trained custom model. If None, default as concatenate model_id and '.pt'
        :type model_name: string
        :param zip_file_name: str =None
            Optional, file name for zip file. if None, default as concatenate model_id and '.zip'
        :type zip_file_name: string
        :param add_apache_license:
            Optional, whether to add a Apache-2.0 license file to model zip file
        :type add_apache_license: string
        :param verbose:
            optional, use to print more logs. Default as false
        :type verbose: bool
        :return: no return value expected
        :rtype: None
        """
        if model_name is None:
            model_name = str(self.model_id.split("/")[-1] + ".pt")

        if model_path is None:
            model_path = os.path.join(self.folder_path, str(model_name))
        else:
            model_path = os.path.join(model_path, str(model_name))

        if verbose:
            print("model path is: ", model_path)

        if zip_file_name is None:
            zip_file_name = str(self.model_id.split("/")[-1] + ".zip")

        zip_file_path = os.path.join(self.folder_path, zip_file_name)
        zip_file_name_without_extension = zip_file_name.split(".")[0]

        if verbose:
            print("Zip file name without extension: ", zip_file_name_without_extension)

        tokenizer_json_path = os.path.join(self.folder_path, "tokenizer.json")
        print("tokenizer_json_path: ", tokenizer_json_path)

        if not os.path.exists(tokenizer_json_path):
            raise Exception(
                "Cannot find tokenizer.json file, please check at "
                + tokenizer_json_path
            )
        if not os.path.exists(model_path):
            raise Exception(
                "Cannot find model in the model path , please check at " + model_path
            )

        # Create a ZipFile Object
        with ZipFile(str(zip_file_path), "w") as zipObj:
            zipObj.write(model_path, arcname=str(model_name))
            zipObj.write(
                tokenizer_json_path,
                arcname="tokenizer.json",
            )
        if add_apache_license:
            super()._add_apache_license_to_model_zip_file(zip_file_path)

        print("zip file is saved to " + zip_file_path + "\n")

    def save_as_pt(
        self,
        sentences: [str],
        model_id="sentence-transformers/msmarco-distilbert-base-tas-b",
        model_name: str = None,
        save_json_folder_path: str = None,
        model_output_path: str = None,
        zip_file_name: str = None,
        add_apache_license: bool = False,
    ) -> str:
        """
        Download sentence transformer model directly from huggingface, convert model to torch script format,
        zip the model file and its tokenizer.json file to prepare to upload to the Open Search cluster

        :param sentences:
            Required, for example  sentences = ['today is sunny']
        :type sentences: List of string [str]
        :param model_id:
            sentence transformer model id to download model from sentence transformers.
            default model_id = "sentence-transformers/msmarco-distilbert-base-tas-b"
        :type model_id: string
        :param model_name:
            Optional, model name to name the model file, e.g, "sample_model.pt". If None, default takes the
            model_id and add the extension with ".pt"
        :type model_name: string
        :param save_json_folder_path:
             Optional, path to save model json file, e.g, "home/save_pre_trained_model_json/"). If None, default as
             default_folder_path from the constructor
        :type save_json_folder_path: string
        :param model_output_path:
             Optional, path to save traced model zip file. If None, default as
             default_folder_path from the constructor
        :type model_output_path: string
        :param zip_file_name:
            Optional, file name for zip file. e.g, "sample_model.zip". If None, default takes the model_id
            and add the extension with ".zip"
        :type zip_file_name: string
        :param add_apache_license:
            Optional, whether to add a Apache-2.0 license file to model zip file
        :type add_apache_license: string
        :return: model zip file path. The file path where the zip file is being saved
        :rtype: string
        """

        model = SentenceTransformer(model_id)

        if model_name is None:
            model_name = str(model_id.split("/")[-1] + ".pt")

        model_path = os.path.join(self.folder_path, model_name)

        if save_json_folder_path is None:
            save_json_folder_path = self.folder_path

        if model_output_path is None:
            model_output_path = self.folder_path

        if zip_file_name is None:
            zip_file_name = str(model_id.split("/")[-1] + ".zip")
        zip_file_path = os.path.join(model_output_path, zip_file_name)

        # handle when model_max_length is unproperly defined in model's tokenizer (e.g. "intfloat/e5-small-v2")
        # (See PR #219 and https://github.com/huggingface/transformers/issues/14561 for more context)
        if model.tokenizer.model_max_length > model.get_max_seq_length():
            model.tokenizer.model_max_length = model.get_max_seq_length()
            print(
                f"The model_max_length is not properly defined in tokenizer_config.json. Setting it to be {model.tokenizer.model_max_length}"
            )

        # save tokenizer.json in save_json_folder_name
        model.save(save_json_folder_path)
        super()._fill_null_truncation_field(
            save_json_folder_path, model.tokenizer.model_max_length
        )

        # convert to pt format will need to be in cpu,
        # set the device to cpu, convert its input_ids and attention_mask in cpu and save as .pt format
        device = torch.device("cpu")
        cpu_model = model.to(device)
        features = cpu_model.tokenizer(
            sentences, return_tensors="pt", padding=True, truncation=True
        ).to(device)

        compiled_model = torch.jit.trace(
            cpu_model,
            (
                {
                    "input_ids": features["input_ids"],
                    "attention_mask": features["attention_mask"],
                }
            ),
            strict=False,
        )
        torch.jit.save(compiled_model, model_path)
        print("model file is saved to ", model_path)

        # zip model file along with tokenizer.json (and license file) as output
        with ZipFile(str(zip_file_path), "w") as zipObj:
            zipObj.write(
                model_path,
                arcname=str(model_name),
            )
            zipObj.write(
                os.path.join(save_json_folder_path, "tokenizer.json"),
                arcname="tokenizer.json",
            )
        if add_apache_license:
            super()._add_apache_license_to_model_zip_file(zip_file_path)

        self.torch_script_zip_file_path = zip_file_path
        print("zip file is saved to ", zip_file_path, "\n")
        return zip_file_path

    def save_as_onnx(
        self,
        model_id="sentence-transformers/msmarco-distilbert-base-tas-b",
        model_name: str = None,
        save_json_folder_path: str = None,
        model_output_path: str = None,
        zip_file_name: str = None,
        add_apache_license: bool = False,
    ) -> str:
        """
        Download sentence transformer model directly from huggingface, convert model to onnx format,
        zip the model file and its tokenizer.json file to prepare to upload to the Open Search cluster

        :param model_id:
            sentence transformer model id to download model from sentence transformers.
            default model_id = "sentence-transformers/msmarco-distilbert-base-tas-b"
        :type model_id: string
        :param model_name:
            Optional, model name to name the model file, e.g, "sample_model.pt". If None, default takes the
            model_id and add the extension with ".pt"
        :type model_name: string
        :param save_json_folder_path:
             Optional, path to save model json file, e.g, "home/save_pre_trained_model_json/"). If None, default as
             default_folder_path from the constructor
        :type save_json_folder_path: string
        :param model_output_path:
             Optional, path to save traced model zip file. If None, default as
             default_folder_path from the constructor
        :type model_output_path: string
        :param zip_file_name:
            Optional, file name for zip file. e.g, "sample_model.zip". If None, default takes the model_id
            and add the extension with ".zip"
        :type zip_file_name: string
        :param add_apache_license:
            Optional, whether to add a Apache-2.0 license file to model zip file
        :type add_apache_license: string
        :return: model zip file path. The file path where the zip file is being saved
        :rtype: string
        """

        model = SentenceTransformer(model_id)

        if model_name is None:
            model_name = str(model_id.split("/")[-1] + ".onnx")

        model_path = os.path.join(self.folder_path, "onnx", model_name)

        if save_json_folder_path is None:
            save_json_folder_path = self.folder_path

        if model_output_path is None:
            model_output_path = self.folder_path

        if zip_file_name is None:
            zip_file_name = str(model_id.split("/")[-1] + ".zip")

        zip_file_path = os.path.join(model_output_path, zip_file_name)

        # handle when model_max_length is unproperly defined in model's tokenizer (e.g. "intfloat/e5-small-v2")
        # (See PR #219 and https://github.com/huggingface/transformers/issues/14561 for more context)
        if model.tokenizer.model_max_length > model.get_max_seq_length():
            model.tokenizer.model_max_length = model.get_max_seq_length()
            print(
                f"The model_max_length is not properly defined in tokenizer_config.json. Setting it to be {model.tokenizer.model_max_length}"
            )

        # save tokenizer.json in output_path
        model.save(save_json_folder_path)
        super()._fill_null_truncation_field(
            save_json_folder_path, model.tokenizer.model_max_length
        )

        convert(
            framework="pt",
            model=model_id,
            output=Path(model_path),
            opset=15,
        )

        print("model file is saved to ", model_path)

        # zip model file along with tokenizer.json (and license file) as output
        with ZipFile(str(zip_file_path), "w") as zipObj:
            zipObj.write(
                model_path,
                arcname=str(model_name),
            )
            zipObj.write(
                os.path.join(save_json_folder_path, "tokenizer.json"),
                arcname="tokenizer.json",
            )
        if add_apache_license:
            super()._add_apache_license_to_model_zip_file(zip_file_path)

        self.onnx_zip_file_path = zip_file_path
        print("zip file is saved to ", zip_file_path, "\n")
        return zip_file_path

    def set_up_accelerate_config(
        self,
        compute_environment: str = None,
        num_machines: int = 1,
        num_processes: int = None,
        verbose: bool = False,
    ) -> None:
        """
        Get default config setting based on the number of GPU on the machine
        if users require other configs, users can run !acclerate config for more options

        :param compute_environment:
            optional, compute environment type to run model, if None, default using 'LOCAL_MACHINE'
        :type compute_environment: string
        :param num_machines:
            optional, number of machine to run model , if None, default using 1
        :type num_machines: int
        :param num_processes:
            optional, number of processes to run model, if None, default to check how many gpus are available and
            use all. if no gpu is available, use cpu
        :type num_processes: int
        :param verbose:
            optional, use printing more logs. Default as false
        :type verbose: bool
        :return: no return value expected
        :rtype: None
        """

        if compute_environment is None or compute_environment == 0:
            compute_environment = "LOCAL_MACHINE"
        else:
            subprocess.run("accelerate config")
            return

        hf_cache_home = os.path.expanduser(
            os.getenv(
                "HF_HOME",
                os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "huggingface"),
            )
        )

        cache_dir = os.path.join(hf_cache_home, "accelerate")

        file_path = os.path.join(cache_dir, "default_config.yaml")
        use_cpu = False
        if verbose:
            print("generated config file: at " + file_path + "\n")
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        if num_processes is None:
            if torch.cuda.is_available():
                num_processes = torch.cuda.device_count()
            else:
                num_processes = 1
                use_cpu = True

        model_config_content = [
            {
                "compute_environment": compute_environment,
                "deepspeed_config": {
                    "gradient_accumulation_steps": 1,
                    "offload_optimizer_device": "none",
                    "offload_param_device": "none",
                    "zero3_init_flag": False,
                    "zero_stage": 2,
                },
                "distributed_type": "DEEPSPEED",
                "downcast_bf16": "no",
                "fsdp_config": {},
                "machine_rank": 0,
                "main_process_ip": None,
                "main_process_port": None,
                "main_training_function": "main",
                "mixed_precision": "no",
                "num_machines": num_machines,
                "num_processes": num_processes,
                "use_cpu": use_cpu,
            }
        ]

        if verbose:
            print("Printing model config content: \n")
            print(model_config_content)

        try:
            with open(file_path, "w") as file:
                if verbose:
                    print(
                        "generating config file for ml common upload: "
                        + file_path
                        + "\n"
                    )
                yaml.dump(model_config_content, file)
        except IOError:
            print(
                "Failed to open config file for ml common upload: " + file_path + "\n"
            )

    def _get_model_description_from_readme_file(self, readme_file_path) -> str:
        """
        Get description of the model from README.md file in the model folder
        after the model is saved in local directory

        See example here:
        https://huggingface.co/sentence-transformers/msmarco-distilbert-base-tas-b/blob/main/README.md)

        This function assumes that the README.md has the following format:

        # sentence-transformers/msmarco-distilbert-base-tas-b
        This is [ ... further description ... ]

        # [ ... Next section ...]
        ...

        :param readme_file_path: Path to README.md file
        :type readme_file_path: string
        :return: Description of the model
        :rtype: string
        """
        readme_data = MarkDownFile.read_file(readme_file_path)

        # Find the description section
        start_str = f"\n# {self.model_id}"
        start = readme_data.find(start_str)
        if start == -1:
            model_name = self.model_id.split("/")[1]
            start_str = f"\n# {model_name}"
            start = readme_data.find(start_str)
        end = readme_data.find("\n#", start + len(start_str))

        # If we cannot find the scope of description section, raise error.
        if start == -1 or end == -1:
            assert False, "Cannot find description in README.md file"

        # Parse out the description section
        description = readme_data[start + len(start_str) + 1 : end].strip()
        description = description.split("\n")[0]

        # Remove hyperlink and reformat text
        description = re.sub(r"\(.*?\)", "", description)
        description = re.sub(r"[\[\]]", "", description)
        description = re.sub(r"\*", "", description)

        # Remove unnecessary part if exists (i.e. " For an introduction to ...")
        # (Found in https://huggingface.co/sentence-transformers/multi-qa-mpnet-base-dot-v1/blob/main/README.md)
        unnecessary_part = description.find(" For an introduction to")
        if unnecessary_part != -1:
            description = description[:unnecessary_part]

        return description

    def _generate_default_model_description(self, embedding_dimension) -> str:
        """
        Generate default model description of the model based on embedding_dimension

        ::param embedding_dimension: Embedding dimension of the model.
        :type embedding_dimension: int
        :return: Description of the model
        :rtype: string
        """
        print(
            "Using default description from embedding_dimension instead (You can overwrite this by specifying description parameter in make_model_config_json function"
        )
        description = f"This is a sentence-transformers model: It maps sentences & paragraphs to a {embedding_dimension} dimensional dense vector space."
        return description

    def make_model_config_json(
        self,
        model_name: str = None,
        version_number: str = 1,
        model_format: str = "TORCH_SCRIPT",
        model_zip_file_path: str = None,
        embedding_dimension: int = None,
        pooling_mode: str = None,
        normalize_result: bool = None,
        description: str = None,
        all_config: str = None,
        additional_config: dict = None,
        model_type: str = None,
        verbose: bool = False,
    ) -> str:
        """
        Parse from config.json file of pre-trained hugging-face model to generate a ml-commons_model_config.json file.
        If all required fields are given by users, use the given parameters and will skip reading the config.json

        :param model_name:
            Optional, The name of the model. If None, default is model id, for example,
            'sentence-transformers/msmarco-distilbert-base-tas-b'
        :type model_name: string
        :param model_format:
            Optional, the format of the model. Default is "TORCH_SCRIPT".
        :type model_format: string
        :param model_zip_file_path:
            Optional, path to the model zip file. Default is the zip file path used in save_as_pt or save_as_onnx
            depending on model_format. This zip file is used to compute model_content_size_in_bytes and
            model_content_hash_value.
        :type model_zip_file_path: string
        :param version_number:
            Optional, The version number of the model. Default is 1
        :type version_number: string
        :param embedding_dimension: Optional, the embedding dimension of the model. If None, get embedding_dimension
            from the pre-trained hugging-face model object.
        :type embedding_dimension: int
        :param pooling_mode: Optional, the pooling mode of the model. If None, get pooling_mode
            from the pre-trained hugging-face model object.
        :type pooling_mode: string
        :param normalize_result: Optional, whether to normalize the result of the model. If None, check from the pre-trained
        hugging-face model object.
        :type normalize_result: bool
        :param description: Optional, the description of the model. If None, get description from the README.md
            file in the model folder.
        :type description: str
        :param all_config:
            Optional, the all_config of the model. If None, parse all contents from the config file of pre-trained
            hugging-face model
        :type all_config: str
        :param additional_config:
            Optional, the additional_config of the model. If None, set additional_config as an empty dictionary
        :type additional_config: dict
        :param model_type:
            Optional, the model_type of the model. If None, parse model_type from the config file of pre-trained
            hugging-face model
        :type model_type: string
        :param verbose:
            optional, use printing more logs. Default as false
        :type verbose: bool
        :return: model config file path. The file path where the model config file is being saved
        :rtype: string
        """
        folder_path = self.folder_path
        config_json_file_path = os.path.join(folder_path, "config.json")
        if model_name is None:
            model_name = self.model_id

        # if user input model_type/embedding_dimension/pooling_mode, it will skip this step.
        model = SentenceTransformer(self.model_id)
        if (
            model_type is None
            or embedding_dimension is None
            or pooling_mode is None
            or normalize_result is None
        ):
            try:
                if embedding_dimension is None:
                    embedding_dimension = model.get_sentence_embedding_dimension()

                for str_idx, module in model._modules.items():
                    if model_type is None and isinstance(module, Transformer):
                        model_type = module.auto_model.__class__.__name__
                        model_type = model_type.lower().rstrip("model")
                    elif pooling_mode is None and isinstance(module, Pooling):
                        pooling_mode = module.get_pooling_mode_str().upper()
                    elif normalize_result is None and isinstance(module, Normalize):
                        normalize_result = True
                    # TODO: Support 'Dense' module
                if normalize_result is None:
                    normalize_result = False
            except Exception as e:
                raise Exception(
                    f"Raised exception while getting model data from pre-trained hugging-face model object: {e}"
                )

        if description is None:
            readme_file_path = os.path.join(self.folder_path, "README.md")
            if os.path.exists(readme_file_path):
                try:
                    if verbose:
                        print("reading README.md file")
                    description = self._get_model_description_from_readme_file(
                        readme_file_path
                    )
                except Exception as e:
                    print(f"Cannot scrape model description from README.md file: {e}")
                    description = self._generate_default_model_description(
                        embedding_dimension
                    )
            else:
                print("Cannot find README.md file to scrape model description")
                description = self._generate_default_model_description(
                    embedding_dimension
                )

        if all_config is None:
            if not os.path.exists(config_json_file_path):
                raise Exception(
                    str(
                        "Cannot find config.json in"
                        + config_json_file_path
                        + ". Please check the config.son file in the path."
                    )
                )
            try:
                with open(config_json_file_path) as f:
                    if verbose:
                        print("reading config file from: " + config_json_file_path)
                    config_content = json.load(f)
                    if all_config is None:
                        all_config = config_content
            except IOError:
                print(
                    "Cannot open in config.json file at ",
                    config_json_file_path,
                    ". Please check the config.json ",
                    "file in the path.",
                )
        if additional_config is None:
            additional_config = {}
        if not additional_config.get("space_type"):
            model_name_suffix = self.model_id.split("/")[-1]
            if model_name_suffix not in self.MODEL_SPACE_TYPE_MAPPING:
                print(
                    f"Default space type cannot be determined for model '{model_name_suffix}'. Consider adding space_type in additional_config."
                )
            else:
                space_type = self.MODEL_SPACE_TYPE_MAPPING[model_name_suffix]
                additional_config["space_type"] = space_type

        model_config_content = {
            "name": model_name,
            "version": version_number,
            "description": description,
            "model_format": model_format,
            "function_name": "TEXT_EMBEDDING",
            "model_config": {
                "model_type": model_type,
                "embedding_dimension": embedding_dimension,
                "framework_type": "sentence_transformers",
                "pooling_mode": pooling_mode,
                "normalize_result": normalize_result,
                "all_config": json.dumps(all_config),
            },
        }
        if additional_config:
            model_config_content["model_config"][
                "additional_config"
            ] = additional_config

        if model_zip_file_path is None:
            model_zip_file_path = (
                self.torch_script_zip_file_path
                if model_format == "TORCH_SCRIPT"
                else self.onnx_zip_file_path
            )
            if model_zip_file_path is None:
                print(
                    "The model configuration JSON file currently lacks the 'model_content_size_in_bytes' and 'model_content_hash_value' fields. You can include these fields by specifying the 'model_zip_file_path' parameter. Failure to do so may result in the model registration process encountering issues."
                )
            else:
                model_config_content["model_content_size_in_bytes"] = os.stat(
                    model_zip_file_path
                ).st_size
                model_config_content["model_content_hash_value"] = (
                    _generate_model_content_hash_value(model_zip_file_path)
                )

        if verbose:
            print("generating ml-commons_model_config.json file...\n")
            print(json.dumps(model_config_content, indent=4))

        model_config_file_path = os.path.join(
            folder_path, "ml-commons_model_config.json"
        )
        os.makedirs(os.path.dirname(model_config_file_path), exist_ok=True)
        with open(model_config_file_path, "w") as file:
            json.dump(model_config_content, file, indent=4)
        print(
            "ml-commons_model_config.json file is saved at : ", model_config_file_path
        )

        return model_config_file_path

    # private methods
    def __qryrem(self, x):
        # for removing the "QRY:" token if they exist in passages
        if "QRY:" in x.passages and "<|startoftext|>" in x.passages:
            y = x.passages.split(" QRY:")[0].split("<|startoftext|>")
            return y[1]
        else:
            return x

    def __is_notebook(self) -> bool:
        try:
            shell = get_ipython().__class__.__name__
            # Jupyter notebook or qtconsole
            if shell == "ZMQInteractiveShell":
                return True
                # Terminal running IPython
            elif shell == "TerminalInteractiveShell":
                return False
            else:
                return False
        except NameError:
            # Probably standard Python interpreter
            return False



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import configparser
import logging
import sys
import termios
import tty
from typing import Optional
from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError
from colorama import Fore, Style, init
from opensearchpy import OpenSearch, RequestsHttpConnection

from opensearch_py_ml.ml_commons.cli.aws_config import AWSConfig
from opensearch_py_ml.ml_commons.cli.cli_base import CLIBase
from opensearch_py_ml.ml_commons.cli.opensearch_domain_config import (
    OpenSearchDomainConfig,
)

# Initialize colorama for colored terminal output
init(autoreset=True)

# Configure the logger for this module
logger = logging.getLogger(__name__)


class PasswordInputError(Exception):
    """Exception for password input errors."""

    pass


class Setup(CLIBase):
    """
    Handles the setup and configuration of the OpenSearch environment.
    Manages AWS credentials (if amazon-opensearch-service) and OpenSearch client initialization.
    """

    def __init__(self):
        """
        Initialize the Setup class with default or existing configurations.
        """
        super().__init__()
        self.config = configparser.ConfigParser()
        self.service_type = ""
        self.ssl_check_enabled = True
        self.opensearch_client = None
        self.session = None
        self.opensearch_config = OpenSearchDomainConfig(
            opensearch_domain_region="",
            opensearch_domain_name="",
            opensearch_domain_username="",
            opensearch_domain_password="",
            opensearch_domain_endpoint="",
        )
        self.aws_config = AWSConfig(
            aws_user_name="",
            aws_role_name="",
            aws_access_key="",
            aws_secret_access_key="",
            aws_session_token="",
        )

    def check_credentials_validity(
        self,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        session_token: Optional[str] = None,
        use_config: bool = False,
    ) -> bool:
        """
        Check if the provided AWS credentials are valid.

        Args:
            access_key (optional): AWS access key ID. Defaults to None.
            secret_key (optional): Aws secret access key. Defaults to None.
            session_token (optional): AWS session token. Defaults to None.
            use_config (optional): Whether to use credentials from config file instead of providede parameters. Defaults to False.

        Returns:
            bool: True if credentials are valid, False otherwise
        """
        try:
            if use_config:
                # Check credentials from a config file
                aws_credentials = self.config.get("aws_credentials", {})
                credentials = {
                    "aws_access_key_id": aws_credentials.get("aws_access_key", ""),
                    "aws_secret_access_key": aws_credentials.get(
                        "aws_secret_access_key", ""
                    ),
                    "aws_session_token": aws_credentials.get("aws_session_token", ""),
                }
            else:
                credentials = {
                    "aws_access_key_id": access_key,
                    "aws_secret_access_key": secret_key,
                    "aws_session_token": session_token,
                }
            self.session = boto3.Session(**credentials)
            sts_client = self.session.client("sts")
            sts_client.get_caller_identity()
            return True
        except ClientError:
            return False

    def update_aws_credentials(
        self, access_key: str, secret_key: str, session_token: str
    ) -> None:
        """
        Update AWS credentials in the config file.

        Args:
            access_key: AWS access key ID to store.
            secret_key: AWS secret access key to store.
            session_token: AWS session token to store.

        Raises:
            Exception: If updating credentials fails, with error details in log
        """
        try:
            if "aws_credentials" not in self.config:
                self.config["aws_credentials"] = {}

            self.config["aws_credentials"]["aws_access_key"] = access_key
            self.config["aws_credentials"]["aws_secret_access_key"] = secret_key
            self.config["aws_credentials"]["aws_session_token"] = session_token
        except Exception as e:
            logger.error(
                f"{Fore.RED}Failed to update AWS credentials: {e}{Style.RESET_ALL}"
            )
            raise

    def check_and_configure_aws(self, config_path: str) -> None:
        """
        Check if AWS credentials are configured and offer to reconfigure if needed.

        Args:
            config_path: Path to the configuration file to update.
        """
        if not self.check_credentials_validity(use_config=True):
            logger.warning(
                f"{Fore.YELLOW}Your AWS credentials are invalid or have expired.{Style.RESET_ALL}"
            )
            self.configure_aws()
            self.config["aws_credentials"][
                "aws_access_key"
            ] = self.aws_config.aws_access_key
            self.config["aws_credentials"][
                "aws_secret_access_key"
            ] = self.aws_config.aws_secret_access_key
            self.config["aws_credentials"][
                "aws_session_token"
            ] = self.aws_config.aws_session_token
            self.update_config(self.config, config_path)
        else:
            print(
                f"{Fore.GREEN}AWS credentials are already configured.{Style.RESET_ALL}"
            )
            reconfigure = input("Do you want to reconfigure? (yes/no): ").lower()
            if reconfigure == "yes":
                self.configure_aws()

    def configure_aws(self) -> None:
        """
        Configure AWS credentials by prompting the user for input.
        """
        print("\nLet's configure your AWS credentials.")
        self.aws_config.aws_access_key = self.get_password_with_asterisks(
            "Enter your AWS Access Key ID: "
        )
        self.aws_config.aws_secret_access_key = self.get_password_with_asterisks(
            "Enter your AWS Secret Access Key: "
        )
        self.aws_config.aws_session_token = self.get_password_with_asterisks(
            "Enter your AWS Session Token: "
        )

        self.update_aws_credentials(
            self.aws_config.aws_access_key,
            self.aws_config.aws_secret_access_key,
            self.aws_config.aws_session_token,
        )
        if self.check_credentials_validity(
            self.aws_config.aws_access_key,
            self.aws_config.aws_secret_access_key,
            self.aws_config.aws_session_token,
        ):
            print(
                f"{Fore.GREEN}New AWS credentials have been successfully configured and verified.{Style.RESET_ALL}"
            )
        else:
            logger.warning(
                f"{Fore.RED}The provided credentials are invalid or expired.{Style.RESET_ALL}"
            )

    def get_password_with_asterisks(
        self, prompt: str = "Enter password: "
    ) -> Optional[str]:
        """
        Get password input from user, masking it with asterisks.

        Args:
            prompt (optional): The prompt text to display before password input. Defaults to "Enter password: ".

        Returns:
            Optional[str]:
                - str: The entered password if input successful
                - None: If interrupted or error occurs
        """

        def _get_windows_password() -> str:
            try:
                import msvcrt

                password = ""
                while True:
                    key = msvcrt.getch()
                    if key == b"\r":  # Enter
                        sys.stdout.write("\n")
                        return password
                    elif key == b"\x08":  # Backspace
                        if password:
                            password = password[:-1]
                            sys.stdout.write("\b \b")
                            sys.stdout.flush()
                    elif key == b"\x03":  # Ctrl+C
                        raise KeyboardInterrupt
                    else:
                        try:
                            char = key.decode("utf-8")
                            password += char
                            sys.stdout.write("*")
                            sys.stdout.flush()
                        except UnicodeDecodeError:
                            continue
            except Exception as e:
                raise PasswordInputError(
                    f"Error reading Windows password input: {str(e)}"
                )

        def _get_unix_password() -> str:
            fd = sys.stdin.fileno()
            old_settings = None
            try:
                old_settings = termios.tcgetattr(fd)
                tty.setraw(fd)
                sys.stdout.write(prompt)
                sys.stdout.flush()

                password = ""
                while True:
                    ch = sys.stdin.read(1)
                    if ch in ("\r", "\n"):  # Enter
                        sys.stdout.write("\r\n")
                        return password
                    elif ch == "\x7f":  # Backspace
                        if password:
                            password = password[:-1]
                            sys.stdout.write("\b \b")
                            sys.stdout.flush()
                    elif ch == "\x03":  # Ctrl+C
                        raise KeyboardInterrupt
                    else:
                        password += ch
                        sys.stdout.write("*")
                        sys.stdout.flush()

            except Exception as e:
                raise PasswordInputError(f"Error reading Unix password input: {str(e)}")
            finally:
                if old_settings is not None:
                    termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

        try:
            if sys.platform == "win32":
                return _get_windows_password()
            return _get_unix_password()
        except KeyboardInterrupt:
            print("\nPassword input interrupted")
            return None
        except PasswordInputError as e:
            print(f"\nError: {str(e)}")
            return None

    def setup_configuration(self) -> str:
        """
        Set up the configuration by prompting the user for various settings.

        Returns:
            str: Path to the created configuration file.
        """
        # Prompt for service type
        print("\nChoose OpenSearch service type:")
        print("1. Amazon OpenSearch Service")
        print("2. Open-source")
        service_choice = input("Enter your choice (1-2): ").strip()

        if service_choice == "1":
            self.service_type = self.AMAZON_OPENSEARCH_SERVICE
        elif service_choice == "2":
            self.service_type = self.OPEN_SOURCE
        else:
            logger.warning(
                f"\n{Fore.YELLOW}Invalid choice. Defaulting to 'amazon-opensearch-service'.{Style.RESET_ALL}"
            )
            self.service_type = self.AMAZON_OPENSEARCH_SERVICE

        # Based on service type, prompt for different configurations
        if self.service_type == self.AMAZON_OPENSEARCH_SERVICE:
            print("\n--- Amazon OpenSearch Service Setup ---")
            self.configure_aws()

            # Prompt for ARN type
            print("\nChoose ARN type:")
            print("1. IAM Role ARN")
            print("2. IAM User ARN")
            arn_type = input("Enter your choice (1-2): ").strip()

            if arn_type == "1":
                self.aws_config.aws_role_name = (
                    input("Enter your AWS IAM Role ARN: ").strip()
                    or self.aws_config.aws_role_name
                )
            elif arn_type == "2":
                self.aws_config.aws_user_name = (
                    input("Enter your AWS IAM User ARN: ").strip()
                    or self.aws_config.aws_user_name
                )
            else:
                logger.warning(
                    f"\n{Fore.YELLOW}Invalid choice. Defaulting to 'IAM Role ARN'.{Style.RESET_ALL}"
                )
                self.aws_config.aws_role_name = (
                    input("Enter your AWS IAM Role ARN: ").strip()
                    or self.aws_config.aws_role_name
                )

            # Prompt for domain information
            default_region = "us-west-2"
            self.opensearch_config.opensearch_domain_region = (
                input(
                    f"Enter your AWS OpenSearch region, or press Enter for default [{default_region}]: "
                ).strip()
                or default_region
            )
            self.opensearch_config.opensearch_domain_endpoint = input(
                "Enter your AWS OpenSearch domain endpoint: "
            ).strip()
            self.opensearch_config.opensearch_domain_username = input(
                "Enter your AWS OpenSearch username: "
            ).strip()
            self.opensearch_config.opensearch_domain_password = (
                self.get_password_with_asterisks("Enter your AWS OpenSearch password: ")
            )
        elif self.service_type == self.OPEN_SOURCE:
            # For open-source, skip AWS configurations
            print("\n--- Open-source OpenSearch Setup ---")
            default_endpoint = "https://localhost:9200"
            self.opensearch_config.opensearch_domain_endpoint = (
                input(
                    f"\nEnter your custom endpoint, or press Enter for default [{default_endpoint}]: "
                ).strip()
                or default_endpoint
            )
            if self.opensearch_config.opensearch_domain_endpoint.startswith("https://"):
                ssl_check = (
                    input(
                        "Do you want to disable SSL certificate check? (yes/no), or press Enter for default 'Enable': "
                    )
                    .strip()
                    .lower()
                    or self.ssl_check_enabled
                )
                if ssl_check == "yes":
                    self.ssl_check_enabled = False
            auth_required = (
                input(
                    "Does your OpenSearch instance require authentication? (yes/no): "
                )
                .strip()
                .lower()
            )
            if auth_required == "yes":
                self.opensearch_config.opensearch_domain_username = input(
                    "Enter your OpenSearch username: "
                ).strip()
                self.opensearch_config.opensearch_domain_password = (
                    self.get_password_with_asterisks("Enter your OpenSearch password: ")
                )
            else:
                self.opensearch_config.opensearch_domain_username = None
                self.opensearch_config.opensearch_domain_password = None

            # AWS OpenSearch region and IAM principal not needed
            self.opensearch_config.opensearch_domain_region = ""
            self.aws_config.aws_role_name = ""
            self.aws_config.aws_user_name = ""

        # Update configuration dictionary
        self.config = {
            "service_type": self.service_type,
            "ssl_check_enabled": self.ssl_check_enabled,
            "opensearch_config": {
                "opensearch_domain_region": self.opensearch_config.opensearch_domain_region,
                "opensearch_domain_endpoint": self.opensearch_config.opensearch_domain_endpoint,
                "opensearch_domain_username": self.opensearch_config.opensearch_domain_username,
                "opensearch_domain_password": self.opensearch_config.opensearch_domain_password,
            },
            "aws_credentials": {
                "aws_role_name": self.aws_config.aws_role_name,
                "aws_user_name": self.aws_config.aws_user_name,
                "aws_access_key": self.aws_config.aws_access_key,
                "aws_secret_access_key": self.aws_config.aws_secret_access_key,
                "aws_session_token": self.aws_config.aws_session_token,
            },
        }
        config_path = self.save_yaml_file(self.config, "configuration", False)
        return config_path

    def initialize_opensearch_client(self) -> bool:
        """
        Initialize the OpenSearch client based on the service type and configuration.

        Returns:
            bool: True if client initialization successful, False otherwise.
        """
        if not self.opensearch_config.opensearch_domain_endpoint:
            logger.warning(
                f"{Fore.RED}OpenSearch endpoint not set. Please run setup first.{Style.RESET_ALL}\n"
            )
            return False

        parsed_url = urlparse(self.opensearch_config.opensearch_domain_endpoint)

        # Determine auth based on service type
        if self.service_type == self.AMAZON_OPENSEARCH_SERVICE:
            if (
                not self.opensearch_config.opensearch_domain_username
                or not self.opensearch_config.opensearch_domain_password
            ):
                logger.warning(
                    f"{Fore.RED}OpenSearch username or password not set. Please run setup first.{Style.RESET_ALL}\n"
                )
                return False
            auth = (
                self.opensearch_config.opensearch_domain_username,
                self.opensearch_config.opensearch_domain_password,
            )
        elif self.service_type == self.OPEN_SOURCE:
            if (
                self.opensearch_config.opensearch_domain_username
                and self.opensearch_config.opensearch_domain_password
            ):
                auth = (
                    self.opensearch_config.opensearch_domain_username,
                    self.opensearch_config.opensearch_domain_password,
                )
            else:
                auth = None
        else:
            logger.warning("Invalid service type. Please check your configuration.")
            return False

        try:
            self.opensearch_client = OpenSearch(
                hosts=[self.opensearch_config.opensearch_domain_endpoint],
                http_auth=auth,
                use_ssl=(parsed_url.scheme == "https"),
                verify_certs=self.ssl_check_enabled,
                ssl_show_warn=False,
                connection_class=RequestsHttpConnection,
                pool_maxsize=20,
            )
            print(
                f"{Fore.GREEN}Initialized OpenSearch client with endpoint: {self.opensearch_config.opensearch_domain_endpoint}{Style.RESET_ALL}\n"
            )
            return True
        except Exception as e:
            logger.error(
                f"{Fore.RED}Error initializing OpenSearch client: {e}{Style.RESET_ALL}\n"
            )
            return False

    def _update_from_config(self) -> bool:
        """
        Update instance variables from loaded config dictionary.
        """
        if not self.config:
            return False

        self.service_type = self.config.get("service_type", "")
        self.ssl_check_enabled = self.config.get("ssl_check_enabled")
        # OpenSearch config
        opensearch_config = self.config.get("opensearch_config", {})
        self.opensearch_config.opensearch_domain_region = opensearch_config.get(
            "opensearch_domain_region", ""
        )
        self.opensearch_config.opensearch_domain_endpoint = opensearch_config.get(
            "opensearch_domain_endpoint", ""
        )
        self.opensearch_config.opensearch_domain_username = opensearch_config.get(
            "opensearch_domain_username", ""
        )
        self.opensearch_config.opensearch_domain_password = opensearch_config.get(
            "opensearch_domain_password", ""
        )
        # AWS credentials
        aws_credentials = self.config.get("aws_credentials", {})
        self.aws_config.aws_role_name = aws_credentials.get("aws_role_name", "")
        self.aws_config.aws_user_name = aws_credentials.get("aws_user_name", "")
        self.aws_config.aws_access_key = aws_credentials.get("aws_access_key", "")
        self.aws_config.aws_secret_access_key = aws_credentials.get(
            "aws_secret_access_key", ""
        )
        self.aws_config.aws_session_token = aws_credentials.get("aws_session_token", "")
        return True

    def _process_config(self, config_path: str) -> Optional[str]:
        """
        Process the configuration file.
        """
        if self.load_config(config_path):
            if self._update_from_config():
                if self.service_type == self.AMAZON_OPENSEARCH_SERVICE:
                    self.check_and_configure_aws(config_path)
                print(
                    f"{Fore.GREEN}\nSetup complete. You are now ready to use the ML features.{Style.RESET_ALL}"
                )
                return config_path
            else:
                logger.warning(
                    f"{Fore.RED}Failed to update configuration.{Style.RESET_ALL}"
                )
        else:
            logger.warning(
                f"{Fore.YELLOW}Could not load existing configuration. Creating new configuration...{Style.RESET_ALL}"
            )
            config_path = self.setup_configuration()

    def setup_command(self, config_path: Optional[str] = None) -> Optional[str]:
        """
        Main setup command that orchestrates the entire setup process.

        Args:
            config_path (optional): Path to existing configuration file. Defaults to None.

        Returns:
            Optional[str]:
                - str: Path to the configuration file if setup successful.
                - None: If setup failed.
        """
        # Check if setup config file path is given in the command
        if not config_path:
            config_exist = (
                input("\nDo you already have a configuration file? (yes/no): ")
                .strip()
                .lower()
            )
            if config_exist == "yes":
                print("\nGreat! Let's use your existing configuration.")
                config_path = input(
                    "Enter the path to your existing configuration file: "
                ).strip()
                if config_path == "":
                    print(
                        f"\n{Fore.YELLOW}No configuration file path provided. Exiting.{Style.RESET_ALL}"
                    )
                    return None
                return self._process_config(config_path)
            else:
                print("Let's create a new configuration file.")
                config_path = self.setup_configuration()

            # Initialize OpenSearch client
            if config_path:
                if self.initialize_opensearch_client():
                    print(
                        f"{Fore.GREEN}Setup complete. You are now ready to use the ML features.{Style.RESET_ALL}"
                    )
                    return config_path
                else:
                    # Handle failure to initialize OpenSearch client
                    logger.warning(
                        f"\n{Fore.RED}Failed to initialize OpenSearch client. Setup incomplete.{Style.RESET_ALL}\n"
                    )
                    return None

        return self._process_config(config_path)



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
import logging
from typing import Any, Dict, Optional

from colorama import Fore, Style, init
from rich.console import Console

from opensearch_py_ml.ml_commons.cli.cli_base import CLIBase

# Initialize Rich console for enhanced CLI outputs
console = Console()

# Initialize colorama for colored terminal output
init(autoreset=True)

# Configure the logger for this module
logger = logging.getLogger(__name__)


class ModelManager(CLIBase):
    """
    Handles the model operations.
    """

    def __init__(self):
        """
        Initialize the ModelManager class.
        """
        super().__init__()
        self.config = {}

    def initialize_predict_model(
        self,
        config_path: str,
        model_id: Optional[str] = None,
        body: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """
        Orchestrates the entire model prediction process.

        Args:
            config_path: Path to the setup configuration file.
            model_id (optional): The model ID to use for prediction.
            body (optional): Prediction request body.

        Returns:
            bool: True if prediction successful, False otherwise
        """
        try:
            # Load and check configuration
            config_result = self.load_and_check_config(config_path)
            if not config_result:
                return False
            ai_helper, _, _, _ = config_result

            # Prompt for model id and predict request body if not provided
            if not model_id:
                model_id = input("\nEnter the model ID: ").strip()
            if not body:
                print(
                    "\nEnter your predict request body as a JSON object (press Enter twice when done): "
                )
                json_input = ""
                lines = []
                while True:
                    line = input()
                    if line.strip() == "":
                        break
                    lines.append(line)

                json_input = "\n".join(lines)
                body = json.loads(json_input)
            else:
                body = json.loads(body)

            response, status = ai_helper.predict(model_id, body)
            if status == 200:
                print(f"{Fore.GREEN}\nSuccessfully predict the model.{Style.RESET_ALL}")

                predict_output = input(
                    "Do you want to save the full prediction output? (yes/no): "
                ).lower()
                if predict_output == "yes":
                    self.predict_model_output(model_id, response)
                return True
            else:
                logger.warning(f"{Fore.RED}Failed to predict model.{Style.RESET_ALL}")
                return False
        except Exception as e:
            logger.error(f"{Fore.RED}Error predicting model: {str(e)}{Style.RESET_ALL}")
            return False

    def initialize_register_model(
        self,
        config_path: str,
        connector_id: Optional[str] = None,
        model_name: Optional[str] = None,
        model_description: Optional[str] = None,
    ) -> bool:
        """
        Orchestrates the entire model registration process.

        Args:
            config_path: Path to the setup configuration file.
            connector_id (optional): The connector ID to register the model with.
            model_name (optional): Name of the model.
            model_description (optional): Description of the model.

        Returns:
            bool: True if registration successful, False otherwise
        """
        try:
            # Load and check configuration
            config_result = self.load_and_check_config(config_path)
            if not config_result:
                return False
            ai_helper, _, _, _ = config_result

            # Prompt for model name, description, and connector ID if not provided
            if not model_name:
                model_name = input("\nEnter the model name: ").strip()
            if not model_description:
                model_description = input("Enter the model description: ").strip()
            if not connector_id:
                connector_id = input("Enter the connector ID: ").strip()

            model_id = ai_helper.register_model(
                model_name, model_description, connector_id
            )

            if model_id:
                print(
                    f"{Fore.GREEN}\nSuccessfully registered a model with ID: {model_id}{Style.RESET_ALL}"
                )
                self.register_model_output(model_id, model_name, connector_id)
                return True
            else:
                logger.warning(f"{Fore.RED}Failed to register model.{Style.RESET_ALL}")
                return False
        except Exception:
            return False



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.
import json
import os
import re
from typing import Optional
from zipfile import ZipFile

import torch
from transformers import AutoModelForMaskedLM, AutoTokenizer

from opensearch_py_ml.ml_commons.ml_common_utils import (
    SPARSE_ENCODING_FUNCTION_NAME,
    _generate_model_content_hash_value,
)
from opensearch_py_ml.ml_models.base_models import SparseModel

# l0 activation. apply the log(1+relu(x)) operator twice in activation process.
ACTIVATION_L0 = "l0"


def _generate_default_model_description() -> str:
    """
    Generate default model description

    :return: Description of the model
    :rtype: string
    """
    print(
        "Using default description (You can overwrite this by specifying description parameter in "
        "make_model_config_json function"
    )
    description = "This is a neural sparse encoding model: It transfers text into sparse vector, and then extract nonzero index and value to entry and weights. It serves only in ingestion and customer should use tokenizer model in query."
    return description


def _sanitize_module_name(name: str) -> str:
    name = re.sub(r"[^0-9A-Za-z_\.]", "_", name)
    parts = []
    for p in name.split("."):
        if not p:
            continue
        if p[0].isdigit():
            p = f"n_{p}"
        parts.append(p)
    return ".".join(parts)


def sanitize_model_modules(model: torch.nn.Module) -> None:
    seen: set[type] = set()
    for m in model.modules():
        cls = m.__class__
        if cls in seen:
            continue
        safe = _sanitize_module_name(getattr(cls, "__module__", ""))
        if safe and safe != cls.__module__:
            cls.__module__ = safe
        seen.add(cls)


class SparseEncodingModel(SparseModel):
    """
    Class for  exporting and configuring the NeuralSparseV2Model model.
    """

    DEFAULT_MODEL_ID = "opensearch-project/opensearch-neural-sparse-encoding-v1"

    def __init__(
        self,
        model_id: str = DEFAULT_MODEL_ID,
        folder_path: str = None,
        overwrite: bool = False,
        sparse_prune_ratio: float = 0,
        activation: str = None,
        model_init_kwargs: Optional[dict] = None,
    ) -> None:

        super().__init__(model_id, folder_path, overwrite)
        if model_init_kwargs is None:
            model_init_kwargs = {}
        self.tokenizer = AutoTokenizer.from_pretrained(model_id, **model_init_kwargs)
        self.backbone_model = AutoModelForMaskedLM.from_pretrained(
            model_id, _attn_implementation="eager", **model_init_kwargs
        )
        default_folder_path = os.path.join(
            os.getcwd(), "opensearch_neural_sparse_model_files"
        )
        if folder_path is None:
            self.folder_path = default_folder_path
        else:
            self.folder_path = folder_path

        if os.path.exists(self.folder_path) and not overwrite:
            print(
                "To prevent overwriting, please enter a different folder path or delete the folder or enable "
                "overwrite = True "
            )
            raise Exception(
                str("The default folder path already exists at : " + self.folder_path)
            )
        self.model_id = model_id
        self.torch_script_zip_file_path = None
        self.onnx_zip_file_path = None
        self.sparse_prune_ratio = sparse_prune_ratio
        self.activation = activation

    def save_as_pt(
        self,
        sentences: [str],
        model_id=DEFAULT_MODEL_ID,
        model_name: str = None,
        save_json_folder_path: str = None,
        model_output_path: str = None,
        zip_file_name: str = None,
        add_apache_license: bool = True,
    ) -> str:
        """
        Download sparse encoding model directly from huggingface, convert model to torch script format,
        zip the model file and its tokenizer.json file to prepare to upload to the OpenSearch cluster

        :param sentences:
            Required, for example  sentences = ['today is sunny']
        :type sentences: List of string [str]
        :param model_id:
             model id to download model from a sparse encoding model.
            default model_id = "opensearch-project/opensearch-neural-sparse-encoding-v1"
        :type model_id: string
        :param model_name:
            Optional, model name to name the model file, e.g, "sample_model.pt". If None, default takes the
            model_id and add the extension with ".pt"
        :type model_name: string
        :param save_json_folder_path:
             Optional, path to save model json file, e.g, "home/save_pre_trained_model_json/"). If None, default as
             default_folder_path from the constructor
        :type save_json_folder_path: string
        :param model_output_path:
             Optional, path to save traced model zip file. If None, default as
             default_folder_path from the constructor
        :type model_output_path: string
        :param zip_file_name:
            Optional, file name for zip file. e.g, "sample_model.zip". If None, default takes the model_id
            and add the extension with ".zip"
        :type zip_file_name: string
        :param add_apache_license:
            Optional, whether to add Apache-2.0 license file to model zip file
        :type add_apache_license: string
        :return: model zip file path. The file path where the zip file is being saved
        :rtype: string
        """
        if model_name is None:
            model_name = str(model_id.split("/")[-1] + ".pt")

        model_path = os.path.join(self.folder_path, model_name)

        if save_json_folder_path is None:
            save_json_folder_path = self.folder_path

        if model_output_path is None:
            model_output_path = self.folder_path

        if zip_file_name is None:
            zip_file_name = str(model_id.split("/")[-1] + ".zip")
        zip_file_path = os.path.join(model_output_path, zip_file_name)

        model = NeuralSparseModel(
            self.backbone_model,
            self.tokenizer,
            self.sparse_prune_ratio,
            self.activation,
        )

        # save tokenizer.json in save_json_folder_name
        self.tokenizer.save_pretrained(save_json_folder_path)

        super()._fill_null_truncation_field(
            save_json_folder_path, self.tokenizer.model_max_length
        )

        # convert to pt format will need to be in cpu,
        # set the device to cpu, convert its input_ids and attention_mask in cpu and save as .pt format
        device = torch.device("cpu")
        cpu_model = model.to(device)

        features = self.tokenizer(
            sentences,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.tokenizer.model_max_length,
            return_attention_mask=True,
            return_token_type_ids=False,
            return_tensors="pt",
        ).to(device)

        sanitize_model_modules(cpu_model)
        compiled_model = torch.jit.trace(cpu_model, dict(features), strict=False)
        torch.jit.save(compiled_model, model_path)
        print("model file is saved to ", model_path)

        # zip model file along with self.tokenizer.json (and license file) as output
        with ZipFile(str(zip_file_path), "w") as zipObj:
            zipObj.write(
                model_path,
                arcname=str(model_name),
            )
            zipObj.write(
                os.path.join(save_json_folder_path, "tokenizer.json"),
                arcname="tokenizer.json",
            )
        if add_apache_license:
            super()._add_apache_license_to_model_zip_file(zip_file_path)

        self.torch_script_zip_file_path = zip_file_path
        print("zip file is saved to ", zip_file_path, "\n")
        return zip_file_path

    def save_as_onnx(self, *args, **kwargs):
        raise NotImplementedError

    def make_model_config_json(
        self,
        model_name: str = None,
        version_number: str = "1.0.0",
        model_format: str = "TORCH_SCRIPT",
        description: str = None,
        model_zip_file_path: str = None,
    ) -> str:
        folder_path = self.folder_path
        if model_name is None:
            model_name = self.model_id

        model_config_content = {
            "name": model_name,
            "version": version_number,
            "model_format": model_format,
            "function_name": SPARSE_ENCODING_FUNCTION_NAME,
        }
        if model_zip_file_path is None:
            model_zip_file_path = (
                self.torch_script_zip_file_path
                if model_format == "TORCH_SCRIPT"
                else self.onnx_zip_file_path
            )
            if model_zip_file_path is None:
                raise Exception(
                    "The model configuration JSON file currently lacks the 'model_content_size_in_bytes' and "
                    "'model_content_hash_value' fields. You can include these fields by specifying the "
                    "'model_zip_file_path' parameter. Failure to do so may result in the model registration process "
                    "encountering issues."
                )
            else:
                model_config_content["model_content_size_in_bytes"] = os.stat(
                    model_zip_file_path
                ).st_size
                model_config_content["model_content_hash_value"] = (
                    _generate_model_content_hash_value(model_zip_file_path)
                )
        if description is not None and description.strip() != "":
            model_config_content["description"] = description
        else:
            model_config_content["description"] = _generate_default_model_description()

        model_config_file_path = os.path.join(
            folder_path, "ml-commons_model_config.json"
        )
        os.makedirs(os.path.dirname(model_config_file_path), exist_ok=True)
        with open(model_config_file_path, "w") as file:
            json.dump(model_config_content, file, indent=4)
        print(
            "ml-commons_model_config.json file is saved at : ", model_config_file_path
        )
        return model_config_file_path

    def get_backbone_model(self):
        if self.backbone_model is not None:
            return self.backbone_model
        else:
            # spda attention is not supported by OpenSearch djl version
            return AutoModelForMaskedLM.from_pretrained(
                self.model_id, _attn_implementation="eager"
            )

    def get_model(self):
        return NeuralSparseModel(
            self.get_backbone_model(),
            self.get_tokenizer(),
            self.sparse_prune_ratio,
            self.activation,
        )

    def save(self, path):
        backbone_model = self.get_backbone_model()
        tokenizer = self.get_tokenizer()
        backbone_model.save_pretrained(path)
        tokenizer.save_pretrained(path)

    def post_process(self):
        pass

    def pre_process(self):
        pass

    def get_tokenizer(self):
        if self.tokenizer is not None:
            return self.tokenizer
        else:
            return AutoTokenizer.from_pretrained(self.model_id)

    def process_sparse_encoding(self, queries):
        return self.get_model().process_sparse_encoding(queries)

    def init_tokenizer(self, model_id=None):
        if model_id is None:
            model_id = self.model_id
        self.tokenizer = AutoTokenizer.from_pretrained(model_id)


INPUT_ID_KEY = "input_ids"
ATTENTION_MASK_KEY = "attention_mask"
OUTPUT_KEY = "output"


class NeuralSparseModel(torch.nn.Module):
    """
    A PyTorch module for transforming input text to sparse vector representation using a pre-trained internal BERT model.
    This class encapsulates the BERT model and provides methods to process text queries into sparse vectors,
    which are easier to handle in sparse data scenarios such as information retrieval.
    """

    def __init__(
        self, backbone_model, tokenizer=None, sparse_prune_ratio=0, activation=None
    ):
        super().__init__()
        self.backbone_model = backbone_model
        self.sparse_prune_ratio = sparse_prune_ratio
        if tokenizer is not None:
            self.tokenizer = tokenizer
            self.special_token_ids = [
                tokenizer.vocab[token]
                for token in tokenizer.special_tokens_map.values()
            ]
            self.id_to_token = ["" for _ in range(len(tokenizer.vocab))]
            for token, idx in tokenizer.vocab.items():
                self.id_to_token[idx] = token
        if activation is None:
            self.activation = lambda values: torch.log(1 + torch.relu(values))
        elif activation == ACTIVATION_L0:
            self.activation = lambda values: torch.log(
                1 + torch.log(1 + torch.relu(values))
            )
        else:
            raise NotImplementedError("Not supported activation function.")

    def forward(self, input: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        result = self.backbone_model(
            input_ids=input[INPUT_ID_KEY], attention_mask=input[ATTENTION_MASK_KEY]
        )[0]
        values, _ = torch.max(result * input[ATTENTION_MASK_KEY].unsqueeze(-1), dim=1)
        values = self.activation(values)
        values[:, self.special_token_ids] = 0
        # max values prune
        # we first find the max values for each sample in batch, then multiply the max value with the sparse_prune_ratio
        # then we only keep those larger than the threshold
        max_values = values.max(dim=-1)[0].unsqueeze(1) * self.sparse_prune_ratio
        values = values * (values > max_values)
        return {OUTPUT_KEY: values}

    def get_sparse_vector(self, feature):
        output = self.forward(feature)
        values = output[OUTPUT_KEY]
        return values

    def transform_sparse_vector_to_dict(self, sparse_vector):
        all_sparse_dicts = []
        for vector in sparse_vector:
            tokens = [
                self.id_to_token[i]
                for i in torch.nonzero(vector, as_tuple=True)[0].tolist()
            ]
            sparse_dict = {
                token: weight.item()
                for token, weight in zip(
                    tokens, vector[torch.nonzero(vector, as_tuple=True)]
                )
            }
            all_sparse_dicts.append(sparse_dict)
        return all_sparse_dicts

    def process_sparse_encoding(self, queries):
        features = self.tokenizer(
            queries,
            padding=True,
            truncation=True,
            return_tensors="pt",
            return_token_type_ids=False,
        )
        sparse_vector = self.get_sparse_vector(features)
        sparse_dict = self.transform_sparse_vector_to_dict(sparse_vector)
        return sparse_dict



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

from typing import Any, Callable, Dict, Optional

from colorama import Fore, Style

from opensearch_py_ml.ml_commons.cli.ai_connector_helper import AIConnectorHelper
from opensearch_py_ml.ml_commons.cli.ml_models.model_base import ModelBase


class AzureOpenAIModel(ModelBase):

    def _get_connector_body(
        self,
        model_type: str,
        resource_name: str,
        deployment_name: str,
        api_version: str,
        openai_api_key: str,
    ) -> Dict[str, Any]:
        """
        Get the connectory body.
        """
        connector_configs = {
            "1": {
                "name": "Azure OpenAI chat completion connector",
                "description": "Connector for Azure OpenAI chat completion model",
                "model": "gpt-4",
                "request_body": '{ "messages": ${parameters.messages}, "temperature": ${parameters.temperature} }',
                "url": "https://${parameters.endpoint}/openai/deployments/${parameters.deploy-name}/chat/completions?api-version=${parameters.api-version}",
                "parameters": {"temperature": 0.7},
            },
            "2": {
                "name": "Azure OpenAI embedding connector",
                "description": "Connector for Azure OpenAI embedding model",
                "model": "text-embedding-ada-002",
                "request_body": '{ "input": ${parameters.input}}',
                "url": "https://${parameters.endpoint}/openai/deployments/${parameters.deploy-name}/embeddings?api-version=${parameters.api-version}",
                "pre_process_function": "connector.pre_process.openai.embedding",
                "post_process_function": "connector.post_process.openai.embedding",
            },
            "3": "Custom model",
        }

        # Handle custom model or invalid choice
        if (
            model_type not in connector_configs
            or connector_configs[model_type] == "Custom model"
        ):
            if model_type not in connector_configs:
                print(
                    f"\n{Fore.YELLOW}Invalid choice. Defaulting to 'Custom model'.{Style.RESET_ALL}"
                )
            return self.input_custom_model_details()

        config = connector_configs[model_type]

        # Base parameters that all connectors need
        base_parameters = {
            "endpoint": f"{resource_name}.openai.azure.com/",
            "deploy-name": f"{deployment_name}",
            "api-version": f"{api_version}",
            "model": config["model"],
        }

        # Merge with model-specific parameters if any
        parameters = {**base_parameters, **config.get("parameters", {})}

        # Return the connector body
        return {
            "name": config["name"],
            "description": config["description"],
            "version": "1.0",
            "protocol": "http",
            "parameters": parameters,
            "credential": {"openAI_key": openai_api_key},
            "actions": [
                {
                    "action_type": "predict",
                    "method": "POST",
                    "headers": {"api-key": openai_api_key},
                    "url": config["url"],
                    "request_body": config["request_body"],
                    **(
                        {"pre_process_function": config["pre_process_function"]}
                        if "pre_process_function" in config
                        else {}
                    ),
                    **(
                        {"post_process_function": config["post_process_function"]}
                        if "post_process_function" in config
                        else {}
                    ),
                }
            ],
        }

    def create_connector(
        self,
        helper: AIConnectorHelper,
        save_config_method: Callable[[str, Dict[str, Any]], None],
        model_name: Optional[str] = None,
        api_key: Optional[str] = None,
        resource_name: Optional[str] = None,
        deployment_name: Optional[str] = None,
        api_version: Optional[str] = None,
        connector_body: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """
        Create Azure OpenAI connector.

        Args:
            helper: Helper instance for OpenSearch connector operations.
            save_config_method: Method to save connector configuration after creation.
            model_name (optional): Specific Azure OpenAI model name.
            api_key (optional): OpenAI API key.
            resource_name (optional): Azure OpenAI resource name.
            deployment_name (optional): Azure OpenAI deployment name.
            api_version (optional): Azure OpenAI API version.
            connector_body (optional): The connector request body.

        Returns:
            bool: True if connector creation successful, False otherwise.
        """
        # Set trusted connector endpoints for Azure OpenAI
        trusted_endpoint = "^https://.*\\.openai\\.azure\\.com/.*$"
        self.set_trusted_endpoint(helper, trusted_endpoint)

        # Prompt to choose model
        model_type = self.get_model_details(
            "Azure OpenAI", self.OPEN_SOURCE, model_name
        )

        # Prompt for API key
        openai_api_key = self.set_api_key(api_key, "OpenAI")

        # Prompt for resource and deployment name, API version
        resource_name = (
            resource_name or input("Enter your Azure OpenAI resource name: ").strip()
        )
        deployment_name = (
            deployment_name
            or input("Enter your Azure OpenAI deployment name: ").strip()
        )
        api_version = (
            api_version or input("Enter your Azure OpenAI API version: ").strip()
        )

        # Get connector body
        connector_body = connector_body or self._get_connector_body(
            model_type, resource_name, deployment_name, api_version, openai_api_key
        )

        # Create connector
        print("\nCreating Azure OpenAI connector...")
        connector_id = helper.create_connector(
            create_connector_role_name=None,
            body=connector_body,
        )

        if connector_id:
            print(
                f"{Fore.GREEN}\nSuccessfully created Azure OpenAI connector with ID: {connector_id}{Style.RESET_ALL}"
            )
            connector_output = helper.get_connector(connector_id)
            save_config_method(connector_id, connector_output)
            return True
        else:
            print(
                f"{Fore.RED}Failed to create Azure OpenAI connector.{Style.RESET_ALL}"
            )
            return False



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.


from typing import Any, Callable, Dict, Optional

from colorama import Fore, Style

from opensearch_py_ml.ml_commons.cli.ai_connector_helper import AIConnectorHelper
from opensearch_py_ml.ml_commons.cli.ml_models.model_base import ModelBase


class BedrockModel(ModelBase):
    def __init__(self, opensearch_domain_region, service_type):
        """
        Initializes the Bedrock model with necessary configurations
        """
        self.opensearch_domain_region = opensearch_domain_region
        self.service_type = service_type

    def _get_connector_body(self, model_type: str, region: str) -> Dict[str, Any]:
        """
        Get the connectory body.
        """
        region_prefix = region.split("-")[0].lower()
        connector_configs = {
            "1": {
                "name": "Amazon Bedrock: AI21 Labs Jurassic-2 Mid",
                "description": "The connector to Bedrock for AI21 Labs Jurassic-2 Mid model",
                "model": "ai21.j2-mid-v1",
                "request_body": '{"prompt":"${parameters.inputs}","maxTokens":200,"temperature":0.7,"topP":1,"stopSequences":[],"countPenalty":{"scale":0},"presencePenalty":{"scale":0},"frequencyPenalty":{"scale":0}}',
                "post_process_function": "\n  return params['completions'][0].data.text; \n",
                "parameters": {},
            },
            "2": {
                "name": "Amazon Bedrock: Anthropic Claude v2",
                "description": "The connector to Bedrock for Claude V2 model",
                "model": "anthropic.claude-v2",
                "request_body": '{"prompt":"\\n\\nHuman: ${parameters.inputs}\\n\\nAssistant:","max_tokens_to_sample":300,"temperature":0.5,"top_k":250,"top_p":1,"stop_sequences":["\\\\n\\\\nHuman:"]}',
                "parameters": {},
            },
            "3": {
                "name": "Amazon Bedrock: Anthropic Claude v3",
                "description": "The connector to Bedrock for Claude V3 model",
                "model": "anthropic.claude-3-sonnet-20240229-v1:0",
                "request_body": '{"messages":[{"role":"user","content":[{"type":"text","text":"${parameters.inputs}"}]}],"anthropic_version":"${parameters.anthropic_version}","max_tokens":${parameters.max_tokens_to_sample}}',
                "parameters": {
                    "auth": "Sig_V4",
                    "response_filter": "$.content[0].text",
                    "max_tokens_to_sample": "8000",
                    "anthropic_version": "bedrock-2023-05-31",
                },
            },
            "4": {
                "name": "Amazon Bedrock: Anthropic Claude v3.7",
                "description": "The connector to Bedrock for Claude V3.7 model",
                "model": f"{region_prefix}.anthropic.claude-3-7-sonnet-20250219-v1:0",
                "request_body": '{ "anthropic_version": "${parameters.anthropic_version}", "max_tokens": ${parameters.max_tokens}, "temperature": ${parameters.temperature}, "messages": ${parameters.messages} }',
                "parameters": {
                    "max_tokens": 8000,
                    "temperature": 1,
                    "anthropic_version": "bedrock-2023-05-31",
                },
            },
            "5": {
                "name": "Amazon Bedrock: Cohere embed-english-v3",
                "description": "The connector to Bedrock for Cohere embed-english-v3 model",
                "model": "cohere.embed-english-v3",
                "request_body": '{ "texts": ${parameters.texts}, "truncate": "${parameters.truncate}", "input_type": "${parameters.input_type}" }',
                "pre_process_function": "connector.pre_process.cohere.embedding",
                "post_process_function": "connector.post_process.cohere.embedding",
                "parameters": {"truncate": "END", "input_type": "search_document"},
            },
            "6": {
                "name": "Amazon Bedrock: Cohere embed-multilingual-v3",
                "description": "The connector to Bedrock for Cohere embed-multilingual-v3",
                "model": "cohere.embed-multilingual-v3",
                "request_body": '{ "texts": ${parameters.texts}, "truncate": "${parameters.truncate}", "input_type": "${parameters.input_type}" }',
                "pre_process_function": "connector.pre_process.cohere.embedding",
                "post_process_function": "connector.post_process.cohere.embedding",
                "parameters": {"truncate": "END", "input_type": "search_document"},
            },
            "7": {
                "name": "Amazon Bedrock: Titan embedding model",
                "description": "The connector to Bedrock for Titan embedding model",
                "model": "amazon.titan-embed-text-v1",
                "request_body": '{ "inputText": "${parameters.inputText}" }',
                "pre_process_function": "connector.pre_process.bedrock.embedding",
                "post_process_function": "connector.post_process.bedrock.embedding",
                "parameters": {},
            },
            "8": {
                "name": "Amazon Bedrock: Titan Mulit-modal model",
                "description": "The connector to Bedrock for Titan Multi-modal model",
                "model": "amazon.titan-embed-image-v1",
                "request_body": '{"inputText": "${parameters.inputText:-null}", "inputImage": "${parameters.inputImage:-null}"}',
                "pre_process_function": "connector.pre_process.bedrock.multimodal_embedding",
                "post_process_function": "connector.post_process.bedrock.embedding",
                "parameters": {"input_docs_processed_step_size": 2},
            },
            "9": "Custom model",
        }

        # Handle custom model or invalid choice
        if (
            model_type not in connector_configs
            or connector_configs[model_type] == "Custom model"
        ):
            if model_type not in connector_configs:
                print(
                    f"\n{Fore.YELLOW}Invalid choice. Defaulting to 'Custom model'.{Style.RESET_ALL}"
                )
            return self.input_custom_model_details()

        config = connector_configs[model_type]

        # Base parameters that all connectors need
        base_parameters = {
            "region": region,
            "service_name": "bedrock",
            "model": config["model"],
        }

        # Merge with model-specific parameters if any
        parameters = {**base_parameters, **config.get("parameters", {})}

        # Return the connector body
        return {
            "name": config["name"],
            "description": config["description"],
            "version": 1,
            "protocol": "aws_sigv4",
            "parameters": parameters,
            "actions": [
                {
                    "action_type": "predict",
                    "method": "POST",
                    "headers": {"content-type": "application/json"},
                    "url": "https://bedrock-runtime.${parameters.region}.amazonaws.com/model/${parameters.model}/invoke",
                    "request_body": config["request_body"],
                    **(
                        {"pre_process_function": config["pre_process_function"]}
                        if "pre_process_function" in config
                        else {}
                    ),
                    **(
                        {"post_process_function": config["post_process_function"]}
                        if "post_process_function" in config
                        else {}
                    ),
                }
            ],
        }

    def _get_connector_role_policy(
        self, model_type: str, connector_body: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Get the IAM role policy for the connector.
        """
        model_name = connector_body["parameters"]["model"]

        # Handle inference profile
        if model_name.startswith(("us.", "eu.")):
            model_arn = [
                f"arn:aws:bedrock:*::foundation-model/{model_name[3:]}",
                f"arn:aws:bedrock:*:*:inference-profile/{model_name}",
            ]
        else:
            model_arn = [f"arn:aws:bedrock:*::foundation-model/{model_name}"]

        return {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": ["bedrock:InvokeModel"],
                    "Effect": "Allow",
                    "Resource": model_arn,
                }
            ],
        }

    def create_connector(
        self,
        helper: AIConnectorHelper,
        save_config_method: Callable[[str, Dict[str, Any]], None],
        connector_role_prefix: Optional[str] = None,
        region: Optional[str] = None,
        model_name: Optional[str] = None,
        connector_body: Optional[Dict[str, Any]] = None,
        aws_access_key: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
    ) -> bool:
        """
        Create Bedrock connector.

        Args:
            helper: Helper instance for OpenSearch connector operations.
            save_config_method: Method to save connector configuration after creation.
            connector_role_prefix (optional): Prefix for role names.
            region (optional): AWS region.
            model_name (optional): Specific Bedrock model name.
            connector_body (optional): The connector request body.
            aws_access_key (optional): AWS access key ID.
            aws_secret_access_key (optional): AWS secet access key.
            aws_session_token (optional): AWS session token.

        Returns:
            bool: True if connector creation successful, False otherwise.
        """
        # Set trusted connector endpoints for Bedrock
        trusted_endpoint = (
            "^https://bedrock-runtime\\..*[a-z0-9-]\\.amazonaws\\.com/.*$"
        )
        self.set_trusted_endpoint(helper, trusted_endpoint)

        # Prompt to choose model
        model_type = self.get_model_details(
            "Amazon Bedrock", self.service_type, model_name
        )

        # Prompt for region
        region = (
            region
            or input(
                f"Enter your AWS region [{self.opensearch_domain_region}]: "
            ).strip()
            or self.opensearch_domain_region
        )

        # Get connector body
        connector_body = connector_body or self._get_connector_body(model_type, region)

        if self.service_type == self.AMAZON_OPENSEARCH_SERVICE:
            # Create connector role
            connector_role_name, create_connector_role_name = (
                self.create_connector_role(connector_role_prefix, "bedrock")
            )
            # Get the connector role inline policy
            connector_role_inline_policy = self._get_connector_role_policy(
                model_type, connector_body
            )

            # Create connector
            print("\nCreating Bedrock connector...")
            connector_id, connector_role_arn, _ = helper.create_connector_with_role(
                connector_role_inline_policy,
                connector_role_name,
                create_connector_role_name,
                connector_body,
                sleep_time_in_seconds=10,
            )
        else:
            # Prompt for AWS credentials
            self.get_aws_credentials(
                connector_body, aws_access_key, aws_secret_access_key, aws_session_token
            )

            # Create connector
            print("\nCreating Bedrock connector...")
            connector_id = helper.create_connector(
                create_connector_role_name=None,
                body=connector_body,
            )

        if connector_id:
            print(
                f"{Fore.GREEN}\nSuccessfully created Bedrock connector with ID: {connector_id}{Style.RESET_ALL}"
            )
            connector_output = helper.get_connector(connector_id)
            save_config_method(
                connector_id,
                connector_output,
                (
                    connector_role_name
                    if self.service_type == self.AMAZON_OPENSEARCH_SERVICE
                    else None
                ),
                (
                    connector_role_arn
                    if self.service_type == self.AMAZON_OPENSEARCH_SERVICE
                    else None
                ),
            )
            return True
        print(f"{Fore.RED}Failed to create Bedrock connector.{Style.RESET_ALL}")
        return False



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
import os
from typing import List
from zipfile import ZipFile

import torch
from torch import nn
from transformers import AutoTokenizer, BertModel, BertPreTrainedModel

from opensearch_py_ml.ml_commons.ml_common_utils import (
    _generate_model_content_hash_value,
)
from opensearch_py_ml.ml_models.base_models import BaseUploadModel

DEFAULT_MODEL_ID = "opensearch-project/opensearch-semantic-highlighter-v1"


class TraceableBertTaggerForSentenceExtractionWithBackoff(BertPreTrainedModel):
    """
    A torch.jit-compatible version of the sentence highlighter model for inference.

    This model extends BERT to perform sentence-level tagging with a backoff mechanism
    that ensures at least one sentence is selected when confidence exceeds a minimum
    threshold (alpha=0.05).
    """

    def __init__(self, config):
        """
        Initialize the model with BERT base and a classification head.

        Parameters
        ----------
        config : BertConfig
            Configuration object containing model hyperparameters
        """
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, self.num_labels)
        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        sentence_ids=None,
    ):
        """
        Forward pass of the model for sentence highlighting.

        Parameters
        ----------
        input_ids : torch.Tensor
            Token IDs of input sequences
        attention_mask : torch.Tensor
            Mask to avoid attention on padding tokens
        token_type_ids : torch.Tensor
            Segment token indices for input portions
        sentence_ids : torch.Tensor
            IDs assigning tokens to sentences

        Returns
        -------
        tuple
            Indices of sentences to highlight for each item
        """
        # Pass inputs through the BERT model
        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
        )

        # Get token-level embeddings and apply dropout
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)

        @torch.jit.script_if_tracing
        def _get_agg_output(ids, sequence_output):
            """
            Aggregates token-level embeddings into sentence-level embeddings.

            Parameters
            ----------
            ids : torch.Tensor
                Tensor containing sentence IDs for each token
            sequence_output : torch.Tensor
                Token-level embeddings from the BERT model

            Returns
            -------
            tuple
                Contains aggregated sentence embeddings, offsets, and sentence counts
            """
            max_sentences = torch.max(ids) + 1
            d_model = sequence_output.shape[-1]

            agg_output = []
            global_offset_per_item = []
            num_sentences_per_item = []
            for i, sen_ids in enumerate(ids):
                out = []
                local_sen_ids = sen_ids.clone()
                mask = local_sen_ids != -100
                offset = local_sen_ids[mask].min()
                global_offset_per_item.append(offset)
                local_sen_ids[mask] = local_sen_ids[mask] - offset
                num_sentences = torch.max(local_sen_ids) + 1
                num_sentences_per_item.append(num_sentences)

                for j in range(int(num_sentences)):
                    out.append(
                        sequence_output[i, local_sen_ids == j].mean(
                            dim=-2, keepdim=True
                        )
                    )
                padding = torch.zeros(
                    (int(max_sentences - num_sentences), d_model),
                    device=sequence_output.device,
                )
                out.append(padding)
                out = torch.cat(out, dim=0)
                agg_output.append(out)
            agg_output = torch.stack(agg_output)
            return (agg_output, global_offset_per_item, num_sentences_per_item)

        # Aggregate token embeddings into sentence embeddings
        agg_output, global_offset_per_item, num_sentences_per_item = _get_agg_output(
            sentence_ids, sequence_output
        )
        logits = self.classifier(agg_output)
        probs = torch.softmax(logits, dim=-1)
        pos_probs = probs[:, :, 1]

        @torch.jit.script_if_tracing
        def _get_sentence_preds(
            pos_probs,
            global_offset_per_item: List[torch.Tensor],
            num_sentences_per_item: List[torch.Tensor],
            threshold: float = 0.5,
            alpha: float = 0.05,
        ):
            """
            Converts sentence probabilities into predictions with backoff logic.

            Parameters
            ----------
            pos_probs : torch.Tensor
                Positive class probabilities for each sentence
            global_offset_per_item : List[torch.Tensor]
                Minimum sentence ID for each batch item
            num_sentences_per_item : List[torch.Tensor]
                Number of sentences for each batch item
            threshold : float, default=0.5
                Probability threshold for sentence selection
            alpha : float, default=0.05
                Minimum confidence threshold for backoff selection

            Returns
            -------
            List[torch.Tensor]
                List of selected sentence indices
            """
            sentences_preds = []
            for pprob, offset, num_sentences in zip(
                pos_probs, global_offset_per_item, num_sentences_per_item
            ):
                # Get probabilities only for valid sentences
                relevant_probs = pprob[:num_sentences]
                # Apply threshold to determine relevant sentences
                relevant_preds = (relevant_probs >= threshold).int()

                # Backoff logic: if no sentence exceeds threshold
                if relevant_preds.sum() == 0:
                    max_prob_idx = relevant_probs.argmax()
                    max_prob = relevant_probs[max_prob_idx].item()
                    if max_prob >= alpha:
                        relevant_preds[max_prob_idx] = 1

                (indices,) = torch.where(relevant_preds == 1)
                indices += offset
                sentences_preds += [indices]

            return sentences_preds

        sentences_preds = _get_sentence_preds(
            pos_probs, global_offset_per_item, num_sentences_per_item
        )

        return tuple(sentences_preds)


class SemanticHighlighterModel(BaseUploadModel):
    """
    Model class for preparing and packaging the OpenSearch semantic highlighter.

    This class handles model conversion to TorchScript, packaging model with tokenizer,
    and generating configuration for OpenSearch ML Commons deployment.
    """

    def __init__(
        self,
        model_id: str = DEFAULT_MODEL_ID,
        folder_path: str = None,
        overwrite: bool = False,
    ) -> None:
        """
        Initialize a SemanticHighlighterModel instance.

        Parameters
        ----------
        model_id : str, optional
            The Hugging Face model ID to use
        folder_path : str, optional
            Directory path to save model files and configuration
        overwrite : bool, optional
            Whether to overwrite existing files
        """
        if folder_path is None:
            folder_path = "semantic-highlighter/"

        super().__init__(
            model_id=model_id, folder_path=folder_path, overwrite=overwrite
        )
        # Path to the generated zip file, populated after calling save_as_pt
        self.torch_script_zip_file_path = None

    def save_as_pt(
        self,
        example_inputs: dict,
        model_id: str = DEFAULT_MODEL_ID,
        model_name: str = None,
        save_json_folder_path: str = None,
        model_output_path: str = None,
        zip_file_name: str = None,
        add_apache_license: bool = True,
    ) -> str:
        """
        Convert model to TorchScript format and prepare it for upload.

        Parameters
        ----------
        example_inputs : dict
            Example inputs for tracing (input_ids, attention_mask, token_type_ids, sentence_ids)
        model_id : str, optional
            Model ID to use from Hugging Face
        model_name : str, optional
            Name for the traced model file
        save_json_folder_path : str, optional
            Path to save config files
        model_output_path : str, optional
            Path to save the traced model
        zip_file_name : str, optional
            Name for the zip file
        add_apache_license : bool, optional
            Whether to add Apache license to the zip file

        Returns
        -------
        str
            Path to the created zip file
        """
        # Generate default model name if not provided
        if model_name is None:
            model_name = str(model_id.split("/")[-1] + ".pt")

        # Create output directories
        os.makedirs(self.folder_path, exist_ok=True)
        model_path = os.path.join(self.folder_path, model_name)

        if save_json_folder_path is None:
            save_json_folder_path = self.folder_path

        if model_output_path is None:
            model_output_path = self.folder_path

        if zip_file_name is None:
            zip_file_name = str(model_id.split("/")[-1] + ".zip")

        zip_file_path = os.path.join(model_output_path, zip_file_name)

        # Auto device selection: prefer GPU if available, fallback to CPU
        target_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        print(f"Using device for tracing: {target_device}")

        # Download and initialize model
        model = TraceableBertTaggerForSentenceExtractionWithBackoff.from_pretrained(
            model_id
        )
        tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased")

        # Move model to target device instead of hardcoded CPU
        model = model.to(target_device)

        # Save tokenizer files
        tokenizer_path = os.path.join(self.folder_path, "tokenizer")
        os.makedirs(tokenizer_path, exist_ok=True)
        tokenizer.save_pretrained(tokenizer_path)
        print(f"Tokenizer files saved to {tokenizer_path}")

        # Move example inputs to the same device as the model
        device_inputs = {}
        for key, tensor in example_inputs.items():
            device_inputs[key] = tensor.to(target_device)
            print(f"Moved {key} to {target_device}: {device_inputs[key].shape}")

        # Trace the model with example inputs on the target device
        traced_model = torch.jit.trace(
            model,
            (
                device_inputs["input_ids"],
                device_inputs["attention_mask"],
                device_inputs["token_type_ids"],
                device_inputs["sentence_ids"],
            ),
        )

        # Move traced model to CPU for saving (standard practice)
        traced_model_cpu = traced_model.cpu()

        # Save the traced model
        torch.jit.save(traced_model_cpu, model_path)
        print(f"Model file saved to {model_path}")

        # Test traced model on both CPU and GPU (if available)
        self._test_traced_model(traced_model_cpu, device_inputs, model_path)

        # Create zip file with model and tokenizer
        with ZipFile(str(zip_file_path), "w") as zipObj:
            zipObj.write(model_path, arcname=str(model_name))

            for file in os.listdir(tokenizer_path):
                file_path = os.path.join(tokenizer_path, file)
                zipObj.write(file_path, arcname=file)

        if add_apache_license:
            super()._add_apache_license_to_model_zip_file(zip_file_path)

        self.torch_script_zip_file_path = zip_file_path
        print(f"Zip file saved to {zip_file_path}")
        return zip_file_path

    def save_as_onnx(
        self,
        example_inputs: dict,
        model_id: str = DEFAULT_MODEL_ID,
        model_name: str = None,
        save_json_folder_path: str = None,
        model_output_path: str = None,
        zip_file_name: str = None,
        add_apache_license: bool = True,
    ) -> str:
        """
        ONNX format is not supported for semantic highlighter models.
        This method is implemented to satisfy the BaseUploadModel interface.

        Raises
        ------
        NotImplementedError
            Always raises this error as ONNX format is not supported.
        """
        raise NotImplementedError(
            "ONNX format is not supported for semantic highlighter models"
        )

    def make_model_config_json(
        self,
        model_name: str = None,
        version_number: str = "1.0.0",
        model_format: str = "TORCH_SCRIPT",
        description: str = None,
        model_zip_file_path: str = None,
    ) -> str:
        """
        Create the model configuration file for OpenSearch ML Commons.

        Parameters
        ----------
        model_name : str, optional
            Name of the model for OpenSearch
        version_number : str, optional
            Version of the model
        model_format : str, optional
            Format of the model
        description : str, optional
            Model description
        model_zip_file_path : str, optional
            Path to the model zip file

        Returns
        -------
        str
            Path to the created config file
        """
        # Use model_id as the model name if none provided
        if model_name is None:
            model_name = self.model_id

        model_config_content = {
            "name": model_name,
            "version": version_number,
            "model_format": model_format,
            "function_name": "QUESTION_ANSWERING",
            "description": (
                description
                if description
                else "A semantic highlighter model that identifies relevant sentences in a document given a query."
            ),
            "model_config": {
                "model_type": "sentence_highlighting",
                "framework_type": "huggingface_transformers",
            },
        }

        if model_zip_file_path is None:
            model_zip_file_path = self.torch_script_zip_file_path

        if model_zip_file_path:
            model_config_content["model_content_size_in_bytes"] = os.stat(
                model_zip_file_path
            ).st_size
            model_config_content["model_content_hash_value"] = (
                _generate_model_content_hash_value(model_zip_file_path)
            )

        model_config_file_path = os.path.join(
            self.folder_path, "ml-commons_model_config.json"
        )
        os.makedirs(os.path.dirname(model_config_file_path), exist_ok=True)

        with open(model_config_file_path, "w") as file:
            json.dump(model_config_content, file, indent=4)

        print(f"Model config file saved at: {model_config_file_path}")
        return model_config_file_path

    def _test_traced_model(self, traced_model_cpu, original_inputs, model_path):
        """
        Test the traced model on both CPU and GPU to ensure compatibility.

        Parameters
        ----------
        traced_model_cpu : torch.jit.ScriptModule
            The traced model on CPU
        original_inputs : dict
            Original inputs used for tracing
        model_path : str
            Path where the model was saved
        """
        print("🧪 Testing traced model compatibility...")

        # Test on CPU
        try:
            loaded_model_cpu = torch.jit.load(
                model_path, map_location=torch.device("cpu")
            )
            cpu_inputs = {k: v.cpu() for k, v in original_inputs.items()}

            cpu_output = loaded_model_cpu(
                cpu_inputs["input_ids"],
                cpu_inputs["attention_mask"],
                cpu_inputs["token_type_ids"],
                cpu_inputs["sentence_ids"],
            )
            print("CPU inference test passed")

        except Exception as e:
            print(f"CPU inference test failed: {e}")
            raise

        # Test on GPU (if available)
        if torch.cuda.is_available():
            try:
                loaded_model_gpu = torch.jit.load(
                    model_path, map_location=torch.device("cuda")
                )
                gpu_inputs = {k: v.cuda() for k, v in original_inputs.items()}

                gpu_output = loaded_model_gpu(
                    gpu_inputs["input_ids"],
                    gpu_inputs["attention_mask"],
                    gpu_inputs["token_type_ids"],
                    gpu_inputs["sentence_ids"],
                )
                print("GPU inference test passed")

                # Compare outputs (move GPU output to CPU for comparison)
                gpu_output_cpu = tuple(tensor.cpu() for tensor in gpu_output)
                if len(cpu_output) == len(gpu_output_cpu):
                    print("CPU and GPU outputs have matching structure")
                else:
                    print("CPU and GPU outputs have different structures")

            except Exception as e:
                print(f"GPU inference test failed: {e}")
                print("Model may not work properly on GPU")
        else:
            print("GPU not available, skipping GPU test")



# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

import json
import logging
import time
import warnings
from typing import Any, Dict, Optional, Tuple
from urllib.parse import urlparse

import boto3
import urllib3
from colorama import Fore, Style
from opensearchpy import OpenSearch, RequestsHttpConnection
from requests_aws4auth import AWS4Auth

from opensearch_py_ml.ml_commons.cli.aws_config import AWSConfig
from opensearch_py_ml.ml_commons.cli.iam_role_helper import IAMRoleHelper
from opensearch_py_ml.ml_commons.cli.opensearch_domain_config import (
    OpenSearchDomainConfig,
)
from opensearch_py_ml.ml_commons.cli.secret_helper import SecretHelper

# Disable warnings when verify_certs=False
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
warnings.filterwarnings("ignore", category=UserWarning, module="opensearchpy")

# Configure the logger for this module
logger = logging.getLogger(__name__)


class AIConnectorHelper:
    """
    Helper class for managing AI connectors and models in OpenSearch.
    """

    OPEN_SOURCE = "open-source"
    AMAZON_OPENSEARCH_SERVICE = "amazon-opensearch-service"

    def __init__(
        self,
        service_type: str,
        opensearch_config: OpenSearchDomainConfig,
        aws_config: AWSConfig,
        ssl_check_enabled: bool = True,
    ):
        """
        Initialize the AIConnectorHelper with necessary AWS and OpenSearch configurations.

        Args:
            service_type: Service type to connect to. Either 'open-source' or 'amazon-opensearch-service'.
            opensearch_config: Configuration object containing OpenSearch domain settings, including region, domain name, credentials, and endpoint.
            aws_config: Configuration object containins AWS credentials and settings including user name, role name, access key, secret key, and session token.
            ssl_check_enabled (optional): Whether to verify SSL certificates when connecting to OpenSearch. Defaults to True.
        """
        self.service_type = service_type
        self.ssl_check_enabled = ssl_check_enabled
        self.opensearch_config = opensearch_config
        self.aws_config = aws_config

        if self.service_type == self.OPEN_SOURCE:
            domain_endpoint = self.opensearch_config.opensearch_domain_endpoint
            domain_arn = None
        else:
            # Get domain info for AOS
            domain_endpoint, domain_arn = self.get_opensearch_domain_info(
                self.opensearch_config.opensearch_domain_region,
                self.opensearch_config.opensearch_domain_name,
                self.aws_config.aws_access_key,
                self.aws_config.aws_secret_access_key,
                self.aws_config.aws_session_token,
            )
        self.opensearch_domain_arn = domain_arn

        # Parse the OpenSearch domain URL
        parsed_url = urlparse(self.opensearch_config.opensearch_domain_endpoint)

        # Initialize OpenSearch client
        self.opensearch_client = OpenSearch(
            hosts=[self.opensearch_config.opensearch_domain_endpoint],
            http_auth=(
                self.opensearch_config.opensearch_domain_username,
                self.opensearch_config.opensearch_domain_password,
            ),
            use_ssl=(parsed_url.scheme == "https"),
            verify_certs=self.ssl_check_enabled,
            connection_class=RequestsHttpConnection,
        )

        # Initialize helpers for IAM roles and secrets management
        if self.service_type == self.OPEN_SOURCE:
            self.iam_helper = None
            self.secret_helper = None
        else:
            self.iam_helper = IAMRoleHelper(
                opensearch_config=self.opensearch_config, aws_config=self.aws_config
            )

            self.secret_helper = SecretHelper(
                opensearch_config=self.opensearch_config, aws_config=self.aws_config
            )

    @staticmethod
    def get_opensearch_domain_info(
        region: str,
        domain_name: str,
        aws_access_key: str,
        aws_secret_access_key: str,
        aws_session_token: str,
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Retrieve the OpenSearch domain endpoint and ARN based on the domain name and region.

        Args:
            region: AWS region where the OpenSearch domain is located (e.g., 'us-west-2').
            domain_name: OpenSearch domain name to retrieve information for.
            aws_access_key: AWS access key.
            aws_secret_access_key: AWS secret access key.
            aws_session_token: AWS session token.

        Returns:
            Tuple[Optional[str], Optional[str]]: A tuple containing:
                - domain_endpoint: The endpoint URL for the OpenSearch domain.
                  (None if retrieval fails)
                - domain_arn: The ARN (Amazon Resource Name) of the OpenSearch domain.
                  (None if retrieval fails)
        """
        try:
            # Get current credentials
            session = boto3.Session(
                aws_access_key_id=aws_access_key,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
            )
            # Check if credentials are valid
            credentials = session.get_credentials()
            if not credentials:
                logger.error(f"{Fore.RED}No valid credentials found.{Style.RESET_ALL}")
                return None, None

            # Get frozen credentials
            frozen_credentials = credentials.get_frozen_credentials()

            # Create client with explicit credentials
            opensearch_client = boto3.client(
                "opensearch",
                region_name=region,
                aws_access_key_id=frozen_credentials.access_key,
                aws_secret_access_key=frozen_credentials.secret_key,
                aws_session_token=frozen_credentials.token,
            )

            response = opensearch_client.describe_domain(DomainName=domain_name)
            domain_status = response["DomainStatus"]
            domain_endpoint = (
                domain_status.get("Endpoint") or domain_status["Endpoints"]["vpc"]
            )
            domain_arn = domain_status["ARN"]
            return domain_endpoint, domain_arn
        except Exception as e:
            logger.error(f"Error retrieving OpenSearch domain info: {e}")
            return None, None

    def get_ml_auth(self, create_connector_role_name: str) -> AWS4Auth:
        """
        Obtain AWS4Auth credentials for ML API calls using the specified IAM role.

        Args:
            create_connector_role_name: Name of the IAM role.

        Returns:
            AWS4Auth: Authentication object containing temporary credentials.
        """
        # Get role ARN
        create_connector_role_arn = self.iam_helper.get_role_arn(
            create_connector_role_name
        )
        if not create_connector_role_arn:
            raise Exception(f"IAM role '{create_connector_role_name}' not found.")

        # Obtain AWS4Auth with temporary credentials
        temp_credentials = self.iam_helper.assume_role(create_connector_role_arn)
        awsauth = AWS4Auth(
            temp_credentials["credentials"]["AccessKeyId"],
            temp_credentials["credentials"]["SecretAccessKey"],
            self.opensearch_config.opensearch_domain_region,
            "es",
            session_token=temp_credentials["credentials"]["SessionToken"],
        )
        return awsauth

    def register_model(
        self,
        model_name: str,
        description: str,
        connector_id: str,
        deploy: bool = True,
    ) -> str:
        """
        Register a remote model in OpenSearch using ML APIs in opensearch-py.

        Args:
            model_name: The name of the model to register with.
            description: The description of the model to register with.
            connector_id: The connector ID to register the model with.
            deploy (optional): Whether to deploy the model immediately after registration. Defaults to True.

        Returns:
            str: The model ID of the registered model.

        Raises:
            KeyError: If the response doesn't contain expected fields (model_id or task_id)
                or if model_id is missing from task response
            Exception: If there's an error during model registration
        """
        try:
            body = {
                "name": model_name,
                "function_name": "remote",
                "description": description,
                "connector_id": connector_id,
            }
            headers = {"Content-Type": "application/json"}
            deploy_str = str(deploy).lower()

            response = self.opensearch_client.plugins.ml.register_model(
                body=body,
                params={"deploy": deploy_str},
                headers=headers,
            )

            if "model_id" in response:
                return response["model_id"]
            elif "task_id" in response:
                task_response = self.opensearch_client.plugins.ml.get_task(
                    response["task_id"],
                )
                print("Task Response:", json.dumps(task_response))
                if "model_id" in task_response:
                    return task_response["model_id"]
                else:
                    raise KeyError(
                        f"'model_id' not found in task response: {task_response}"
                    )
            elif "error" in response:
                raise Exception(f"Error registering model: {response['error']}")
            else:
                raise KeyError(
                    f"The response does not contain 'model_id' or 'task_id'. Response content: {response}"
                )
        except Exception as e:
            logger.error(
                f"{Fore.RED}Error registering model: {str(e)}{Style.RESET_ALL}"
            )
            raise

    # TODO: Replace with self.opensearch_client.plugins.ml.predict() once available in opensearch-py
    # Current workaround uses transport.perform_request directly
    def predict(self, model_id: str, body: Dict[str, Any]) -> Tuple[str, Optional[str]]:
        """
        Make a prediction using the specified model and input body.

        Args:
            model_id: The model ID to use for prediction.
            body: The request body for prediction.

        Returns:
            Tuple[str, Optional[str]]: A tuple containing:
                - response_text: The complete response from the model as a JSON string
                - status: The status code from the inference results, or None if not available
        """
        headers = {"Content-Type": "application/json"}
        response = self.opensearch_client.transport.perform_request(
            method="POST",
            url=f"/_plugins/_ml/models/{model_id}/_predict",
            body=body,
            headers=headers,
        )
        response_text = json.dumps(response)
        status = response.get("inference_results", [{}])[0].get("status_code")
        print(f"Predict Response: {response_text[:200]}...{response_text[-21:]}")
        return response_text, status

    # TODO: Replace with self.opensearch_client.plugins.ml.get_connector() once available in opensearch-py
    # Current workaround uses transport.perform_request directly
    def get_connector(self, connector_id: str) -> str:
        """
        Get a connector information from the connector ID.

        Args:
            connector_id: The connector ID to get information for.

        Returns:
            str: The connector information as a JSON string
        """
        headers = {"Content-Type": "application/json"}
        response = self.opensearch_client.transport.perform_request(
            method="GET",
            url=f"/_plugins/_ml/connectors/{connector_id}",
            headers=headers,
        )
        return json.dumps(response)

    def create_connector(
        self, create_connector_role_name: str, body: Dict[str, Any]
    ) -> str:
        """
        Create a connector in OpenSearch using the specified role and body.

        Args:
            create_connector_role_name: Name of the IAM role.
            body: The connector configuration as a dictionary.

        Returns:
            str: The ID of the created connector.

        Raises:
            ValueError: If body is None or not a dictionary
        """
        # Verify connector body is not empty and in dict format
        if body is None:
            raise ValueError("A 'body' parameter must be provided as a dictionary.")
        if not isinstance(body, dict):
            raise ValueError("'body' needs to be a dictionary.")

        # Store original auth
        original_auth = None

        # For AOS, temporarily modify client auth
        if self.service_type == self.AMAZON_OPENSEARCH_SERVICE:
            # Obtain AWS4Auth credentials and initialize OpenSearch client with the credentials
            temp_awsauth = self.get_ml_auth(create_connector_role_name)
            original_auth = (
                self.opensearch_client.transport.connection_pool.connections[
                    0
                ].session.auth
            )
            self.opensearch_client.transport.connection_pool.connections[
                0
            ].session.auth = temp_awsauth
        try:
            # Create connector using the body parameter
            headers = {"Content-Type": "application/json"}
            response = self.opensearch_client.plugins.ml.create_connector(
                body=body,
                headers=headers,
            )
            connector_id = response.get("connector_id")
            return connector_id
        finally:
            # Restore original auth if it was modified
            if original_auth is not None:
                self.opensearch_client.http_auth = original_auth

    def _create_iam_role(
        self, step_number: int, connector_role_name: str, inline_policy: Dict[str, Any]
    ) -> str:
        """
        Create an IAM role in OpenSearch using the specified inline policy.
        """
        trust_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": "es.amazonaws.com"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        print(f"Step {step_number}: Create IAM role configured in connector")
        if not self.iam_helper.role_exists(connector_role_name):
            connector_role_arn = self.iam_helper.create_iam_role(
                connector_role_name, trust_policy, inline_policy
            )
        else:
            print(f"{connector_role_name} role exists, skipping creation.")
            connector_role_arn = self.iam_helper.get_role_arn(connector_role_name)
        print("----------")
        return connector_role_arn

    def _configure_iam_role(
        self, step_number: int, connector_role_arn: str, create_connector_role_name: str
    ) -> str:
        """
        Configure an IAM role in OpenSearch.
        """
        statements = []
        if self.aws_config.aws_user_name:
            statements.append(
                {
                    "Effect": "Allow",
                    "Principal": {"AWS": self.aws_config.aws_user_name},
                    "Action": "sts:AssumeRole",
                }
            )
        if self.aws_config.aws_role_name:
            statements.append(
                {
                    "Effect": "Allow",
                    "Principal": {"AWS": self.aws_config.aws_role_name},
                    "Action": "sts:AssumeRole",
                }
            )
        trust_policy = {"Version": "2012-10-17", "Statement": statements}

        inline_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": "iam:PassRole",
                    "Resource": connector_role_arn,
                },
                {
                    "Effect": "Allow",
                    "Action": "es:ESHttpPost",
                    "Resource": self.opensearch_domain_arn,
                },
            ],
        }

        print(f"Step {step_number}: Configure IAM role in OpenSearch")
        print(
            f"Step {step_number}.1: Create IAM role for Signing create connector request"
        )
        if not self.iam_helper.role_exists(create_connector_role_name):
            create_connector_role_arn = self.iam_helper.create_iam_role(
                create_connector_role_name, trust_policy, inline_policy
            )
        else:
            print(f"{create_connector_role_name} role exists, skipping creation.")
            create_connector_role_arn = self.iam_helper.get_role_arn(
                create_connector_role_name
            )
        print("----------")
        return create_connector_role_arn

    def _map_iam_role(
        self,
        step_number: int,
        create_connector_role_arn: str,
        create_connector_role_name: str,
    ) -> None:
        """
        Map IAM role in OpenSearch.
        """
        print(
            f"Step {step_number}.2: Map IAM role {create_connector_role_name} to OpenSearch permission role"
        )
        self.iam_helper.map_iam_role_to_backend_role(create_connector_role_arn)
        print("----------")

    def _create_connector_with_credentials(
        self,
        step_number: int,
        create_connector_input: Dict[str, Any],
        create_connector_role_name: str,
        connector_role_arn: str,
        sleep_time_in_seconds: int,
        secret_arn: Optional[str] = None,
    ) -> Tuple[str, str, Optional[str]]:
        """
        Create connector in OpenSearch with either role ARN only or both secret and role ARN.
        """
        print(f"Step {step_number}: Create connector in OpenSearch")
        print("Waiting for resources to be ready...")
        for remaining in range(sleep_time_in_seconds, 0, -1):
            print(f"\rTime remaining: {remaining} seconds...", end="", flush=True)
            time.sleep(1)
        print("\nWait completed, creating connector...")
        print("Connector role arn: ", connector_role_arn)
        body = create_connector_input
        if secret_arn:
            body["credential"] = {
                "secretArn": secret_arn,
                "roleArn": connector_role_arn,
            }
        else:
            body["credential"] = {"roleArn": connector_role_arn}
        connector_id = self.create_connector(create_connector_role_name, body)
        print("----------")
        return connector_id, connector_role_arn, secret_arn

    def create_connector_with_secret(
        self,
        secret_name: str,
        secret_value: Dict[str, Any],
        connector_role_name: str,
        create_connector_role_name: str,
        create_connector_input: Dict[str, Any],
        sleep_time_in_seconds: int = 10,
    ) -> Tuple[str, str, str]:
        """
        Create a connector in OpenSearch using a secret for credentials.

        Args:
            secret_name: Name for the secret to be created in AWS Secrets Manager.
            secret_value: The secret value to be stored, containing necessary credentials.
            connector_role_name: Name for the IAM role that the connector will use.
            create_connector_role_name: Name for the IAM role used to create the connector.
            create_connector_input: The configuration for the connector.
            sleep_time_in_seconds (optional): Number of seconds to wait before creating the connector. Defaults to 10 seconds.

        Returns:
            Tuple[str, str, str]: A tuple containing:
                - connector_id: The ID of the created connector.
                - connector_role_arn: The ARN of the role used by the connector.
                - secret_arn: The ARN of the created secret.
        """
        # Step 1: Create Secret
        print("Step 1: Create Secret")
        if not self.secret_helper.secret_exists(secret_name):
            secret_arn = self.secret_helper.create_secret(secret_name, secret_value)
        else:
            print(f"{secret_name} secret exists, skipping creation.")
            secret_arn = self.secret_helper.get_secret_arn(secret_name)
        print("----------")

        # Step 2: Create IAM role configured in connector
        inline_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": [
                        "secretsmanager:GetSecretValue",
                        "secretsmanager:DescribeSecret",
                    ],
                    "Effect": "Allow",
                    "Resource": secret_arn,
                }
            ],
        }
        connector_role_arn = self._create_iam_role(
            "2", connector_role_name, inline_policy
        )

        # Step 3: Configure IAM role in OpenSearch
        # 3.1 Create IAM role for signing create connector request
        create_connector_role_arn = self._configure_iam_role(
            "3", connector_role_arn, create_connector_role_name
        )

        # 3.2 Map IAM role to backend role in OpenSearch
        self._map_iam_role("3", create_connector_role_arn, create_connector_role_name)

        # Step 4: Create connector
        return self._create_connector_with_credentials(
            "4",
            create_connector_input,
            create_connector_role_name,
            connector_role_arn,
            sleep_time_in_seconds,
            secret_arn,
        )

    def create_connector_with_role(
        self,
        connector_role_inline_policy: Dict[str, Any],
        connector_role_name: str,
        create_connector_role_name: str,
        create_connector_input: Dict[str, Any],
        sleep_time_in_seconds: int = 10,
    ) -> Tuple[str, str, None]:
        """
        Create a connector in OpenSearch using an IAM role for credentials.

        Args:
            connector_role_inline_policy:
            connector_role_name: Name for the IAM role that the connector will use.
            create_connector_role_name: Name for the IAM role used to create the connector.
            create_connector_input: The configuration for the connector.
            sleep_time_in_seconds (optional): Number of seconds to wait before creating the connector. Defaults to 10 seconds.

        Returns:
            Tuple[str, str, None]: A tuple containing:
                - connector_id (str): The ID of the created connector
                - connector_role_arn (str): The ARN of the role used by the connector
                - None: Placeholder for secret_arn
        """
        # Step 1: Create IAM role configured in connector
        connector_role_arn = self._create_iam_role(
            "1", connector_role_name, connector_role_inline_policy
        )

        # Step 2: Configure IAM role in OpenSearch
        # 2.1 Create IAM role for signing create connector request
        create_connector_role_arn = self._configure_iam_role(
            "2", connector_role_arn, create_connector_role_name
        )

        # 2.2 Map IAM role to backend role in OpenSearch
        self._map_iam_role("2", create_connector_role_arn, create_connector_role_name)

        # Step 3: Create connector
        return self._create_connector_with_credentials(
            "3",
            create_connector_input,
            create_connector_role_name,
            connector_role_arn,
            sleep_time_in_seconds,
        )
